import shutil
import sys
from tempfile import mkdtemp

from kivy.base import EventLoop
from kivy.tests.common import GraphicUnitTest
from kivy.config import ConfigParser
import networkx as nx

from LiSE import Engine
from LiSE.character import Facade
from ELiDE.app import ELiDEApp
from ELiDE.grid.board import GridBoard, GridBoardView
from .util import all_spots_placed, all_pawns_placed, idle_until, \
 window_with_widget, \
 ELiDEAppTest


class GridBoardTest(GraphicUnitTest):

	@staticmethod
	def test_layout_grid():
		spots_wide = 3
		spots_tall = 3
		spot_width = 32
		spot_height = 32
		graph = nx.grid_2d_graph(spots_wide, spots_tall)
		char = Facade(graph)
		something = char.place[1, 1].new_thing('something')
		otherthing = char.place[2, 2].new_thing('otherthing')
		assert len(char.thing) == 2
		board = GridBoard(character=char)
		win = window_with_widget(GridBoardView(board=board))
		while not (all_spots_placed(board, char)
					and all_pawns_placed(board, char)):
			EventLoop.idle()
		otherthing['location'] = board.pawn['otherthing'].loc_name = (0, 0)
		zero = board.spot[0, 0]
		that = board.pawn['otherthing']
		idle_until(lambda: that in zero.children, 1000,
					"pawn 'otherthing' did not relocate within 1000 ticks")
		assert that.parent == zero
		for x in range(spots_wide):
			for y in range(spots_tall):
				spot = board.spot[x, y]
				assert spot.x == x * spot_width
				assert spot.y == y * spot_height
		assert board.pawn['something'].parent == board.spot[1, 1]
		assert board.pawn['something'].pos == board.spot[1, 1].pos
		assert board.pawn['otherthing'].parent == board.spot[0, 0]
		assert board.pawn['otherthing'].pos == board.spot[0, 0].pos


class SwitchGridTest(ELiDEAppTest):

	def test_character_switch_grid(self):
		with Engine(self.prefix) as eng:
			eng.add_character('physical', nx.grid_2d_graph(10, 1))
			eng.add_character('tall', nx.grid_2d_graph(1, 10))
		app = self.app
		window_with_widget(app.build())
		idle_until(
			lambda: hasattr(app, 'mainscreen') and app.mainscreen.mainview and
			app.mainscreen.statpanel and hasattr(app.mainscreen, 'gridview'))
		app.mainscreen.statpanel.toggle_gridview()
		idle_until(lambda: app.mainscreen.gridview in app.mainscreen.mainview.
					children)
		idle_until(lambda: app.mainscreen.gridview.board.children)
		assert all(child.y == 0
					for child in app.mainscreen.gridview.board.children)
		assert not all(child.x == 0
						for child in app.mainscreen.gridview.board.children)
		app.character_name = 'tall'
		idle_until(
			lambda: all(child.x == 0
						for child in app.mainscreen.gridview.board.children),
			1000, "Never got the new board")
		idle_until(
			lambda: not all(child.y == 0 for child in app.mainscreen.gridview.
							board.children), 1000, "New board arranged weird")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.properties import (NumericProperty, ObjectProperty, StringProperty)
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.recycleview import RecycleView
from kivy.lang import Builder
from kivy.graphics import Color, Line


class RuleStepper(RecycleView):
	name = StringProperty()

	def from_rules_handled_turn(self, rules_handled_turn):
		data = [{
			'widget': 'RuleStepperRuleButton',
			'name': 'start',
			'end_tick': 0,
			'height': 40
		}]
		prev_tick = 0
		for rbtyp, rules in sorted(rules_handled_turn.items(),
									key=lambda kv: min(kv[1].keys() or [-1])):
			if not rules:
				continue
			last_entity = None
			last_rulebook = None
			typed = False
			for tick, (entity, rulebook, rule) in sorted(rules.items()):
				if tick == prev_tick:
					continue
				if not typed:
					data.append({'widget': 'RulebookTypeLabel', 'name': rbtyp})
					typed = True
				rulebook_per_entity = rbtyp in {'thing', 'place', 'portal'}
				if not rulebook_per_entity:
					if rulebook != last_rulebook:
						last_rulebook = rulebook
						data.append({
							'widget': 'RulebookLabel',
							'name': rulebook
						})
				if entity != last_entity:
					last_entity = entity
					data.append({'widget': 'EntityLabel', 'name': entity})
				if rulebook_per_entity:
					if rulebook != last_rulebook:
						rulebook = last_rulebook
						data.append({
							'widget': 'RulebookLabel',
							'name': rulebook
						})
				data.append({
					'widget': 'RuleStepperRuleButton',
					'name': rule,
					'start_tick': prev_tick,
					'end_tick': tick,
					'height': 40
				})
				prev_tick = tick
		self.data = data


class RuleStepperRuleButton(Button):
	name = StringProperty()
	start_tick = NumericProperty()
	end_tick = NumericProperty()
	tick = NumericProperty()
	set_tick = ObjectProperty()

	def __init__(self, **kwargs):
		super(RuleStepperRuleButton, self).__init__(**kwargs)
		self.bind(pos=self.upd_line, size=self.upd_line, tick=self.upd_line)

	def on_release(self, *args):
		self.set_tick(self.end_tick)
		self.tick = self.end_tick

	def upd_line(self, *args):
		if hasattr(self, 'color_inst'):
			if self.tick == self.end_tick:
				self.color_inst.rgba = [1, 0, 0, 1]
				self.line.points = [self.x, self.y, self.right, self.y]
			else:
				self.color_inst.rgba = [0, 0, 0, 0]
		else:
			with self.canvas:
				self.color_inst = Color(rgba=([1, 0, 0, 1] if self.tick in (
					self.start_tick, self.end_tick) else [0, 0, 0, 0]))
				self.line = Line(
					points=[self.x, self.top, self.right, self.top])


class EntityLabel(Label):
	name = ObjectProperty()


class RulebookLabel(Label):
	name = ObjectProperty()  # rulebooks may have tuples for names


class RulebookTypeLabel(Label):
	name = StringProperty()


Builder.load_string("""
#:import ScrollEffect kivy.effects.scroll.ScrollEffect
<RuleStepper>:
    key_viewclass: 'widget'
    effect_cls: ScrollEffect
    RecycleGridLayout:
        cols: 1
        size_hint_y: None
        default_size_hint: 1, None
        default_height: 20
        height: self.minimum_height
<RuleStepperRuleButton>:
    text: '\\n'.join((self.name, str(self.end_tick)))
    font_size: 14
    text_size: self.width, None
    halign: 'center'
    tick: app.tick
    set_tick: app.set_tick
<EntityLabel>:
    multiline: True
    text: str(self.name)
    text_size: self.width, None
    size: self.texture_size
    font_size: 14
    padding_x: 8
<RulebookLabel>:
    text: str(self.name)
    text_size: self.width, None
    size: self.texture_size
    font_size: 14
    bold: True
    padding_x: 4
<RulebookTypeLabel>:
    text: self.name
    text_size: self.width, None
    font_size: 16
    bold: True
    size: self.texture_size
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget to represent :class:`Place`s. :class:`Pawn` moves around on
top of these.

"""
from kivy.clock import Clock

from kivy.properties import (AliasProperty, ObjectProperty, NumericProperty)
from .pawnspot import GraphPawnSpot
from ..util import trigger


class GraphSpot(GraphPawnSpot):
	"""The icon that represents a :class:`Place`.

    Each :class:`Spot` is located on the Board that represents the
    :class:`Character` that the underlying :class:`Place` is in. Its
    coordinates are relative to its :class:`Board`, not necessarily
    the window the :class:`Board` is in.

    """
	default_image_paths = ['atlas://rltiles/floor.atlas/floor-stone']
	default_pos = (0.5, 0.5)

	def __init__(self, **kwargs):
		"""Deal with triggers and bindings, and arrange to take care of
        changes in game-time.

        """
		self._pospawn_partials = {}
		self._pospawn_triggers = {}
		kwargs['size_hint'] = (None, None)
		if 'place' in kwargs:
			kwargs['proxy'] = kwargs['place']
			del kwargs['place']
		super().__init__(**kwargs)

	def on_board(self, *args):
		super().on_board(*args)
		self.board.bind(size=self._upd_pos)

	def _upd_pos(self, *args):
		if self.board is None:
			Clock.schedule_once(self._upd_pos, 0)
			return
		self.pos = (int(
			self.proxy.get('_x', self.default_pos[0]) * self.board.width),
					int(
						self.proxy.get('_y', self.default_pos[1]) *
						self.board.height))

	def finalize(self, initial=True):
		if initial:
			self._upd_pos()
		super().finalize(initial)

	def push_pos(self, *args):
		"""Set my current position, expressed as proportions of the graph's
        width and height, into the ``_x`` and ``_y`` keys of the
        entity in my ``proxy`` property, such that it will be
        recorded in the database.

        """
		self.proxy['_x'] = self.x / self.board.width
		self.proxy['_y'] = self.y / self.board.height

	_trigger_push_pos = trigger(push_pos)

	def on_touch_up(self, touch):
		if touch.grab_current is not self:
			return False
		self.center = touch.pos
		self._trigger_push_pos()
		touch.ungrab(self)
		self._trigger_push_pos()
		return True

	def __repr__(self):
		"""Give my name and position."""
		return "<{}@({},{}) at {}>".format(self.name, self.x, self.y, id(self))



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget to display the contents of a :class:`kivy.atlas.Atlas` in
one :class:`kivy.uix.togglebutton.ToggleButton` apiece, arranged in a
:class:`kivy.uix.stacklayout.StackLayout`. The user selects graphics
from the :class:`Pallet`, and the :class:`Pallet` updates its
``selection`` list to show what the user selected."""
from kivy.clock import Clock
from kivy.properties import (DictProperty, NumericProperty, ObjectProperty,
								OptionProperty, ListProperty,
								ReferenceListProperty, StringProperty)
from kivy.resources import resource_find
from kivy.atlas import Atlas
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.stacklayout import StackLayout
from kivy.graphics import Rectangle
from .util import trigger


class SwatchButton(ToggleButton):
	"""Toggle button containing a texture and its name, which, when
    toggled, will report the fact to the :class:`Pallet` it's in.

    """
	tex = ObjectProperty()
	"""Texture to display here"""

	def on_state(self, *args):
		if self.state == 'down':
			assert (self not in self.parent.selection)
			if self.parent.selection_mode == 'single':
				for wid in self.parent.selection:
					if wid is not self:
						wid.state = 'normal'
				self.parent.selection = [self]
			else:
				self.parent.selection.append(self)
		else:
			if self in self.parent.selection:
				self.parent.selection.remove(self)

	def on_parent(self, *args):
		if not self.canvas or not self.tex:
			Clock.schedule_once(self.on_parent, 0)
			return
		self.canvas.after.clear()
		with self.canvas.after:
			self._img_rect = Rectangle(pos=self._get_img_rect_pos(),
										size=self.tex.size,
										texture=self.tex)
		self.fbind('pos', self._upd_img_rect_pos)
		self.fbind('tex', self._upd_img_rect_tex)

	def _get_img_rect_pos(self):
		x, y = self.pos
		width, height = self.size
		tw, th = self.tex.size
		return (x + (width / 2 - tw / 2), y + height - th)

	@trigger
	def _upd_img_rect_pos(self, *args):
		self._img_rect.pos = self._get_img_rect_pos()

	@trigger
	def _upd_img_rect_tex(self, *args):
		self._img_rect.texture = self.tex
		self._img_rect.size = self.tex.size
		self._img_rect.pos = self._get_img_rect_pos()

	def on_size(self, *args):
		self.text_size = self.size


class Pallet(StackLayout):
	"""Many :class:`SwatchButton`, gathered from an :class:`kivy.atlas.Atlas`."""
	atlas = ObjectProperty()
	""":class:`kivy.atlas.Atlas` object I'll make :class:`SwatchButton` from."""
	filename = StringProperty()
	"""Path to an atlas; will construct :class:`kivy.atlas.Atlas` when set"""
	swatches = DictProperty({})
	""":class:`SwatchButton` widgets here, keyed by name of their graphic"""
	swatch_width = NumericProperty(100)
	"""Width of each and every :class:`SwatchButton` here"""
	swatch_height = NumericProperty(75)
	"""Height of each and every :class:`SwatchButton` here"""
	swatch_size = ReferenceListProperty(swatch_width, swatch_height)
	"""Size of each and every :class:`SwatchButton` here"""
	selection = ListProperty([])
	"""List of :class:`SwatchButton`s that are selected"""
	selection_mode = OptionProperty('single', options=['single', 'multiple'])
	"""Whether to allow only a 'single' selected :class:`SwatchButton` (default), or 'multiple'"""

	def on_selection(self, *args):
		Logger.debug('Pallet: {} got selection {}'.format(
			self.filename, self.selection))

	def on_filename(self, *args):
		if not self.filename:
			return
		resource = resource_find(self.filename)
		if not resource:
			raise ValueError("Couldn't find atlas: {}".format(self.filename))
		self.atlas = Atlas(resource)

	def on_atlas(self, *args):
		if self.atlas is None:
			return
		self.upd_textures()
		self.atlas.bind(textures=self._trigger_upd_textures)

	def upd_textures(self, *args):
		"""Create one :class:`SwatchButton` for each texture"""
		if self.canvas is None:
			Clock.schedule_once(self.upd_textures, 0)
			return
		swatches = self.swatches
		atlas_textures = self.atlas.textures
		remove_widget = self.remove_widget
		add_widget = self.add_widget
		swatch_size = self.swatch_size
		for name, swatch in list(swatches.items()):
			if name not in atlas_textures:
				remove_widget(swatch)
				del swatches[name]
		for (name, tex) in atlas_textures.items():
			if name in swatches and swatches[name] != tex:
				remove_widget(swatches[name])
			if name not in swatches or swatches[name] != tex:
				swatches[name] = SwatchButton(text=name,
												tex=tex,
												size_hint=(None, None),
												size=swatch_size)
				add_widget(swatches[name])

	def _trigger_upd_textures(self, *args):
		if hasattr(self, '_scheduled_upd_textures'):
			Clock.unschedule(self._scheduled_upd_textures)
		self._scheduled_upd_textures = Clock.schedule_once(
			self._trigger_upd_textures)


kv = """
<Pallet>:
    orientation: 'lr-tb'
    padding_y: 100
    size_hint: (None, None)
    height: self.minimum_height
"""
Builder.load_string(kv)


class PalletBox(BoxLayout):
	pallets = ListProperty()



from kivy.uix.layout import Layout
from kivy.properties import ObjectProperty

from ..imagestackproxy import ImageStackProxy


class GridSpot(ImageStackProxy, Layout):
	default_image_paths = ['atlas://rltiles/floor.atlas/floor-stone']
	board = ObjectProperty()

	def do_layout(self, *args):
		for child in self.children:
			child.pos = self.pos
			child.size = self.size
			for rect in child._texture_rectangles.values():
				rect.size = self.size

	def add_widget(self, widget, index=0, canvas='after'):
		self._trigger_layout()
		return super().add_widget(widget, index, canvas)



from kivy.tests.common import UnitTestTouch

from LiSE import Engine
from LiSE.examples import polygons
from .util import idle_until, window_with_widget, ELiDEAppTest


class CharacterSwitcherTest(ELiDEAppTest):

	def setUp(self):
		super(CharacterSwitcherTest, self).setUp()
		with Engine(self.prefix) as eng:
			polygons.install(eng)

	def test_character_switcher(self):
		app = self.app
		win = window_with_widget(app.build())
		idle_until(lambda: app.manager.current == 'main', 100,
					'Never switched to main')
		idle_until(lambda: app.mainscreen.boardview, 100,
					'never got boardview')
		idle_until(lambda: app.mainscreen.boardview.board.spot, 100,
					'never got spots')
		physspots = len(app.mainscreen.boardview.board.spot)
		app.mainscreen.charmenu.charmenu.toggle_chars_screen()
		idle_until(lambda: app.manager.current == 'chars', 100,
					'Never switched to chars')
		boxl = app.chars.ids.charsview.ids.boxl
		idle_until(lambda: len(boxl.children) == 3, 100,
					"Didn't get all three characters")
		for charb in boxl.children:
			if charb.text == 'triangle':
				touch = UnitTestTouch(*charb.center)
				touch.pos = charb.center
				assert charb.dispatch('on_touch_down', touch)
				self.advance_frames(5)
				charb.dispatch('on_touch_up', touch)
				idle_until(lambda: charb.state == 'down', 10,
							'Button press did not work')
				break
		else:
			assert False, 'No button for "triangle" character'
		idle_until(
			lambda: app.chars.ids.charsview.character_name == 'triangle', 100,
			"Never propagated character_name")
		app.chars.toggle()
		idle_until(lambda: app.manager.current == 'main', 100,
					"Didn't switch back to main")
		idle_until(
			lambda: not app.mainscreen.boardview.board.spot, 100,
			"Didn't clear out spots, {} left".format(
				len(app.mainscreen.boardview.board.spot)))
		app.mainscreen.charmenu.charmenu.toggle_chars_screen()
		idle_until(lambda: app.manager.current == 'chars', 100,
					'Never switched to chars')
		for charb in boxl.children:
			if charb.text == 'physical':
				touch = UnitTestTouch(*charb.center)
				touch.pos = charb.center
				assert charb.dispatch('on_touch_down', touch)
				self.advance_frames(5)
				charb.dispatch('on_touch_up', touch)
				idle_until(lambda: charb.state == 'down', 10,
							'Button press did not work')
				break
		else:
			assert False, 'No button for "physical" character'
		idle_until(
			lambda: app.chars.ids.charsview.character_name == 'physical', 100,
			'Never propagated character_name')
		app.chars.toggle()
		idle_until(
			lambda: len(app.mainscreen.boardview.board.spot) == physspots, 100,
			"Never got physical back")



from types import SimpleNamespace

from kivy.base import EventLoop
from kivy.config import ConfigParser
from kivy.tests.common import UnitTestTouch

from LiSE.character import Facade
from ELiDE.menu import DirPicker
from ELiDE.screen import MainScreen
from ELiDE.spritebuilder import PawnConfigScreen, SpotConfigScreen
from ELiDE.statcfg import StatScreen
from ELiDE.graph.board import GraphBoard
from ELiDE.grid.board import GridBoard
from .util import ELiDEAppTest, ListenableDict, MockEngine, idle_until, \
 window_with_widget


class MockStore:

	def save(self, *args):
		pass


class ScreenTest(ELiDEAppTest):

	def test_advance_time(self):
		app = self.app
		app.mainmenu = DirPicker()
		app.spotcfg = SpotConfigScreen()
		app.pawncfg = PawnConfigScreen()
		app.statcfg = StatScreen()
		char = Facade()
		char.name = 'physical'
		app.character = char
		app.engine = MockEngine()
		app.strings = MockStore()
		app.funcs = MockStore()
		char.character = SimpleNamespace(engine=app.engine)
		app.engine.character['physical'] = char
		entity = ListenableDict()
		entity.engine = app.engine
		entity.name = 'name'
		app.selected_proxy = app.proxy = app.statcfg.proxy = entity
		screen = MainScreen(
			graphboards={'physical': GraphBoard(character=char)},
			gridboards={'physical': GridBoard(character=char)})
		win = window_with_widget(screen)
		idle_until(lambda: 'timepanel' in screen.ids, 100,
					"timepanel never got id")
		timepanel = screen.ids['timepanel']
		idle_until(lambda: timepanel.size != [100, 100], 100,
					"timepanel never resized")
		turnfield = timepanel.ids['turnfield']
		turn_before = int(turnfield.hint_text)
		stepbut = timepanel.ids['stepbut']
		motion = UnitTestTouch(*stepbut.center)
		motion.touch_down()
		motion.touch_up()
		EventLoop.idle()
		assert int(turnfield.hint_text) == turn_before + 1

	def test_play(self):
		app = self.app
		app.spotcfg = SpotConfigScreen()
		app.pawncfg = PawnConfigScreen()
		app.statcfg = StatScreen()
		app.mainmenu = DirPicker()
		app.strings = MockStore()
		app.funcs = MockStore()
		char = Facade()
		char.name = 'foo'
		app.character = char
		app.engine = MockEngine()
		char.character = SimpleNamespace(engine=app.engine)
		app.engine.character['foo'] = char
		entity = ListenableDict()
		entity.engine = app.engine
		entity.name = 'name'
		app.selected_proxy = app.proxy = app.statcfg.proxy = entity
		screen = MainScreen(graphboards={'foo': GraphBoard(character=char)},
							gridboards={'foo': GridBoard(character=char)},
							play_speed=1.0)
		win = window_with_widget(screen)
		idle_until(lambda: 'timepanel' in screen.ids, 100,
					"timepanel never got id")
		timepanel = screen.ids['timepanel']
		idle_until(lambda: timepanel.size != [100, 100], 100,
					"timepanel never resized")
		turnfield = timepanel.ids['turnfield']
		turn_before = int(turnfield.hint_text)
		playbut = timepanel.ids['playbut']
		motion = UnitTestTouch(*playbut.center)
		motion.touch_down()
		motion.touch_up()
		idle_until(lambda: int(turnfield.hint_text) == 3, 400,
					"Time didn't advance fast enough")



# This file is part of the kivy-garden project.
# Copyright (c) Zachary Spector, public@zacharyspector.com
# Available under the terms of the MIT license.
"""Several textures superimposed on one another, and possibly offset
by some amount.

In 2D games where characters can wear different clothes or hold
different equipment, their graphics are often composed of several
graphics layered on one another. This widget simplifies the management
of such compositions.

"""
from kivy.uix.widget import Widget
from kivy.core.image import Image
from kivy.graphics import (Rectangle, InstructionGroup, PushMatrix, PopMatrix,
							Translate)
from kivy.graphics.fbo import Fbo
from kivy.properties import (AliasProperty, DictProperty, ListProperty,
								ObjectProperty)
from kivy.clock import Clock
from kivy.resources import resource_find


class TextureStack(Widget):
	"""Several textures superimposed on one another, and possibly offset
    by some amount.

    In 2D games where characters can wear different clothes or hold
    different equipment, their graphics are often composed of several
    graphics layered on one another. This widget simplifies the
    management of such compositions.

    """
	texs = ListProperty()
	"""Texture objects"""
	offxs = ListProperty()
	"""x-offsets. The texture at the same index will be moved to the right
    by the number of pixels in this list.

    """
	offys = ListProperty()
	"""y-offsets. The texture at the same index will be moved upward by
    the number of pixels in this list.

    """
	group = ObjectProperty()
	"""My ``InstructionGroup``, suitable for addition to whatever ``canvas``."""

	def _get_offsets(self):
		return zip(self.offxs, self.offys)

	def _set_offsets(self, offs):
		offxs = []
		offys = []
		for x, y in offs:
			offxs.append(x)
			offys.append(y)
		self.offxs, self.offys = offxs, offys

	offsets = AliasProperty(_get_offsets,
							_set_offsets,
							bind=('offxs', 'offys'))
	"""List of (x, y) tuples by which to offset the corresponding texture."""
	_texture_rectangles = DictProperty({})
	"""Private.

    Rectangle instructions for each of the textures, keyed by the
    texture.

    """

	def __init__(self, **kwargs):
		"""Make triggers and bind."""
		kwargs['size_hint'] = (None, None)
		self.translate = Translate(0, 0)
		self.group = InstructionGroup()
		super().__init__(**kwargs)
		self.bind(offxs=self.on_pos, offys=self.on_pos)

	def on_texs(self, *args):
		"""Make rectangles for each of the textures and add them to the canvas."""
		if not self.canvas or not self.texs:
			Clock.schedule_once(self.on_texs, 0)
			return
		texlen = len(self.texs)
		# Ensure each property is the same length as my texs, padding
		# with 0 as needed
		for prop in ('offxs', 'offys'):
			proplen = len(getattr(self, prop))
			if proplen > texlen:
				setattr(self, prop, getattr(self, prop)[:proplen - texlen])
			if texlen > proplen:
				propval = list(getattr(self, prop))
				propval += [0] * (texlen - proplen)
				setattr(self, prop, propval)
		self.group.clear()
		self._texture_rectangles = {}
		w = h = 0
		(x, y) = self.pos
		self.translate.x = x
		self.translate.y = y
		self.group.add(PushMatrix())
		self.group.add(self.translate)
		for tex, offx, offy in zip(self.texs, self.offxs, self.offys):
			rect = Rectangle(pos=(offx, offy), size=tex.size, texture=tex)
			self._texture_rectangles[tex] = rect
			self.group.add(rect)
			tw = tex.width + offx
			th = tex.height + offy
			if tw > w:
				w = tw
			if th > h:
				h = th
		self.size = (w, h)
		self.group.add(PopMatrix())
		self.canvas.add(self.group)

	def on_pos(self, *args):
		"""Translate all the rectangles within this widget to reflect the widget's position.

        """
		(x, y) = self.pos
		self.translate.x = x
		self.translate.y = y

	def clear(self):
		"""Clear my rectangles and ``texs``."""
		self.group.clear()
		self._texture_rectangles = {}
		self.texs = []
		self.size = [1, 1]

	def insert(self, i, tex):
		"""Insert the texture into my ``texs``, waiting for the creation of
        the canvas if necessary.

        """
		if not self.canvas:
			Clock.schedule_once(lambda dt: self.insert(i, tex), 0)
			return
		self.texs.insert(i, tex)

	def append(self, tex):
		"""``self.insert(len(self.texs), tex)``"""
		self.insert(len(self.texs), tex)

	def __delitem__(self, i):
		"""Remove a texture and its rectangle"""
		tex = self.texs[i]
		try:
			rect = self._texture_rectangles[tex]
			self.canvas.remove(rect)
			del self._texture_rectangles[tex]
		except KeyError:
			pass
		del self.texs[i]

	def __setitem__(self, i, v):
		"""First delete at ``i``, then insert there"""
		if len(self.texs) > 0:
			self._no_upd_texs = True
			self.__delitem__(i)
			self._no_upd_texs = False
		self.insert(i, v)

	def pop(self, i=-1):
		"""Delete the texture at ``i``, and return it."""
		return self.texs.pop(i)


class ImageStack(TextureStack):
	"""Instead of supplying textures themselves, supply paths to where the
    textures may be loaded from.

    """
	paths = ListProperty()
	"""List of paths to images you want stacked."""
	pathtexs = DictProperty()
	"""Private. Dictionary mapping image paths to textures of the images."""
	pathimgs = DictProperty()
	"""Dictionary mapping image paths to ``kivy.core.Image`` objects."""

	def on_paths(self, *args):
		"""Make textures from the images in ``paths``, and assign them at the
        same index in my ``texs`` as in my ``paths``.

        """
		for i, path in enumerate(self.paths):
			if path in self.pathtexs:
				if (self.pathtexs[path] in self.texs
					and self.texs.index(self.pathtexs[path]) == i):
					continue
			else:
				self.pathimgs[path] = img = Image.load(resource_find(path),
														keep_data=True)
				self.pathtexs[path] = img.texture
			if i == len(self.texs):
				self.texs.append(self.pathtexs[path])
			else:
				self.texs[i] = self.pathtexs[path]

	def clear(self):
		"""Clear paths, textures, rectangles"""
		self.paths = []
		super().clear()

	def insert(self, i, v):
		"""Insert a string to my paths"""
		if not isinstance(v, str):
			raise TypeError("Paths only")
		self.paths.insert(i, v)

	def __delitem__(self, i):
		"""Delete texture, rectangle, path"""
		super().__delitem__(i)
		del self.paths[i]

	def pop(self, i=-1):
		"""Delete and return a path"""
		r = self.paths[i]
		del self[i]
		return r


class TextureStackBatchWidget(Widget):
	"""Widget for efficiently drawing many TextureStacks

    Only add TextureStack or ImageStack widgets to this. Avoid adding
    any that are to be changed frequently.

    """
	critical_props = ['texs', 'offxs', 'offys', 'pos']
	"""Properties that, when changed on my children, force a redraw."""

	def __init__(self, **kwargs):
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self._trigger_rebind_children = Clock.create_trigger(
			self.rebind_children)
		super(TextureStackBatchWidget, self).__init__(**kwargs)

	def on_parent(self, *args):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		if not hasattr(self, '_fbo'):
			with self.canvas:
				self._fbo = Fbo(size=self.size)
				self._fbo.add_reload_observer(self.redraw)
				self._translate = Translate(x=self.x, y=self.y)
				self._rectangle = Rectangle(texture=self._fbo.texture,
											size=self.size)
		self.rebind_children()

	def rebind_children(self, *args):
		child_by_uid = {}
		binds = {prop: self._trigger_redraw for prop in self.critical_props}
		for child in self.children:
			child_by_uid[child.uid] = child
			child.bind(**binds)
		if hasattr(self, '_old_children'):
			old_children = self._old_children
			for uid in set(old_children).difference(child_by_uid):
				old_children[uid].unbind(**binds)
		self.redraw()
		self._old_children = child_by_uid

	def redraw(self, *args):
		fbo = self._fbo
		fbo.bind()
		fbo.clear()
		fbo.clear_buffer()
		fbo.release()
		for child in self.children:
			assert child.canvas not in fbo.children
			fbo.add(child.canvas)

	def on_pos(self, *args):
		if not hasattr(self, '_translate'):
			return
		self._translate.x, self._translate.y = self.pos

	def on_size(self, *args):
		if not hasattr(self, '_rectangle'):
			return
		self._rectangle.size = self._fbo.size = self.size
		self.redraw()

	def add_widget(self, widget, index=0, canvas=None):
		if not isinstance(widget, TextureStack):
			raise TypeError("TextureStackBatch is only for TextureStack")
		if index == 0 or len(self.children) == 0:
			self.children.insert(0, widget)
		else:
			children = self.children
			if index >= len(children):
				index = len(children)

			children.insert(index, widget)
		widget.parent = self
		if hasattr(self, '_fbo'):
			self.rebind_children()

	def remove_widget(self, widget):
		if widget not in self.children:
			return
		self.children.remove(widget)
		widget.parent = None
		if hasattr(self, '_fbo'):
			self.rebind_children()


if __name__ == '__main__':
	from kivy.base import runTouchApp
	from itertools import cycle
	import json

	# I should come up with a prettier demo that has a whole lot of widgets in it


	class DraggyStack(ImageStack):

		def on_touch_down(self, touch):
			if self.collide_point(*touch.pos):
				touch.grab(self)
				self._old_parent = parent = self.parent
				parent.remove_widget(self)
				parent.parent.add_widget(self)
				assert self in parent.parent.children
				assert self.parent == parent.parent
				return True

		def on_touch_move(self, touch):
			if touch.grab_current is self:
				self.center = touch.pos
				return True

		def on_touch_up(self, touch):
			self.parent.remove_widget(self)
			self.pos = self._old_parent.to_local(*self.pos, relative=True)
			self._old_parent.add_widget(self)
			return True

	with open('marsh_davies_island_bg.atlas') as bgf, open(
		'marsh_davies_island_fg.atlas') as fgf:
		pathses = zip(
			('atlas://marsh_davies_island_bg/' + name for name in json.load(
				bgf)["marsh_davies_island_bg-0.png"].keys()),
			('atlas://marsh_davies_island_fg/' + name for name in cycle(
				json.load(fgf)["marsh_davies_island_fg-0.png"].keys())))
	sbatch = TextureStackBatchWidget(size=(800, 600), pos=(0, 0))
	for i, paths in enumerate(pathses):
		sbatch.add_widget(
			DraggyStack(paths=paths,
						offxs=[0, 16],
						offys=[0, 16],
						pos=(0, 32 * i)))
	parent = Widget(size=(800, 600), pos=(0, 0))
	parent.add_widget(sbatch)
	runTouchApp(parent)



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The big widget that shows the graph of the selected Character."""
from functools import partial
from kivy.properties import (BooleanProperty, ReferenceListProperty,
								DictProperty, ObjectProperty, NumericProperty,
								ListProperty, StringProperty)
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.uix.image import Image
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.scatter import ScatterPlane
from kivy.graphics.transformation import Matrix
from kivy.vector import Vector

from .spot import GraphSpot
from .arrow import GraphArrow, GraphArrowWidget, ArrowLayout, get_points_multi
from .pawn import Pawn
from ..dummy import Dummy
from ..util import trigger
from ..boardview import BoardView
import numpy as np


def normalize_layout(l):
	"""Make sure all the spots in a layout are where you can click.

    Returns a copy of the layout with all spot coordinates are
    normalized to within (0.0, 0.98).

    """
	xs = []
	ys = []
	ks = []
	for (k, (x, y)) in l.items():
		xs.append(x)
		ys.append(y)
		ks.append(k)
	minx = np.min(xs)
	maxx = np.max(xs)
	if maxx == minx:
		xnorm = np.array([0.5] * len(xs))
	else:
		xco = 0.98 / (maxx - minx)
		xnorm = np.multiply(np.subtract(xs, [minx] * len(xs)), xco)
	miny = np.min(ys)
	maxy = np.max(ys)
	if miny == maxy:
		ynorm = np.array([0.5] * len(ys))
	else:
		yco = 0.98 / (maxy - miny)
		ynorm = np.multiply(np.subtract(ys, [miny] * len(ys)), yco)
	return dict(zip(ks, zip(map(float, xnorm), map(float, ynorm))))


class KvLayoutBack(FloatLayout):
	"""What to show behind the graph.

    By default, shows nothing.

    """


class KvLayoutFront(FloatLayout):
	"""What to show in front of the graph.

    By default, shows nothing.

    """
	pass


class FinalLayout(FloatLayout):

	def finalize_all(self, *args):
		for child in self.children:
			child.finalize()
		self.bind(children=self._trigger_finalize_all)

	_trigger_finalize_all = trigger(finalize_all)


class GraphBoard(RelativeLayout):
	"""A graphical view onto a :class:`LiSE.Character`, resembling a game
    graph.

    """
	app = ObjectProperty()
	character = ObjectProperty()
	wallpaper_path = StringProperty()
	spot = DictProperty({})
	pawn = DictProperty({})
	arrow = DictProperty({})
	wallpaper = ObjectProperty()
	kvlayoutback = ObjectProperty()
	arrowlayout = ObjectProperty()
	spotlayout = ObjectProperty()
	kvlayoutfront = ObjectProperty()
	wids = ReferenceListProperty(wallpaper, kvlayoutback, arrowlayout,
									spotlayout, kvlayoutfront)
	spots_unposd = ListProperty([])
	layout_tries = NumericProperty(5)
	tracking_vel = BooleanProperty(False)
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = ObjectProperty(False)
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty(False)
	grabbing = BooleanProperty(True)
	grabbed = ObjectProperty(None, allownone=True)
	spot_cls = ObjectProperty(GraphSpot)
	pawn_cls = ObjectProperty(Pawn)
	arrow_cls = ObjectProperty(GraphArrow, allownone=True)
	proto_arrow_cls = ObjectProperty(GraphArrowWidget)
	_scheduled_rm_spot = DictProperty()
	_scheduled_rm_arrow = DictProperty()
	_scheduled_discard_pawn = DictProperty()
	_scheduled_add_pawn = DictProperty()

	@property
	def widkwargs(self):
		return {'size_hint': (None, None), 'size': self.size, 'pos': (0, 0)}

	def on_touch_down(self, touch):
		"""Check for collisions and select an appropriate entity."""
		if hasattr(self, '_lasttouch') and self._lasttouch == touch:
			return
		if not self.collide_point(*touch.pos):
			return
		touch.push()
		touch.apply_transform_2d(self.to_local)
		if self.app.selection:
			if self.app.selection.collide_point(*touch.pos):
				Logger.debug("Board: hit selection")
				touch.grab(self.app.selection)
		pawns = list(self.pawns_at(*touch.pos))
		if pawns:
			Logger.debug("Board: hit {} pawns".format(len(pawns)))
			self.selection_candidates = pawns
			if self.app.selection in self.selection_candidates:
				self.selection_candidates.remove(self.app.selection)
			touch.pop()
			return True
		spots = list(self.spots_at(*touch.pos))
		if spots:
			Logger.debug("Board: hit {} spots".format(len(spots)))
			self.selection_candidates = spots
			if self.adding_portal:
				self.origspot = self.selection_candidates.pop(0)
				self.protodest = Dummy(name='protodest',
										pos=touch.pos,
										size=(0, 0))
				self.add_widget(self.protodest)
				self.protodest.on_touch_down(touch)
				self.protoportal = self.proto_arrow_cls(
					origin=self.origspot, destination=self.protodest)
				self.add_widget(self.protoportal)
				if self.reciprocal_portal:
					self.protoportal2 = self.proto_arrow_cls(
						destination=self.origspot, origin=self.protodest)
					self.add_widget(self.protoportal2)
			touch.pop()
			return True
		arrows = list(self.arrows_at(*touch.pos))
		if arrows:
			Logger.debug("Board: hit {} arrows".format(len(arrows)))
			self.selection_candidates = arrows
			if self.app.selection in self.selection_candidates:
				self.selection_candidates.remove(self.app.selection)
			if isinstance(
				self.app.selection, GraphArrow
			) and self.app.selection.reciprocal in self.selection_candidates:
				self.selection_candidates.remove(self.app.selection.reciprocal)
			touch.pop()
			return True
		touch.pop()

	def on_touch_move(self, touch):
		"""If an entity is selected, drag it."""
		if hasattr(self, '_lasttouch') and self._lasttouch == touch:
			return
		if self.app.selection in self.selection_candidates:
			self.selection_candidates.remove(self.app.selection)
		if self.app.selection:
			if not self.selection_candidates:
				self.keep_selection = True
			ret = super().on_touch_move(touch)
			return ret
		elif self.selection_candidates:
			for cand in self.selection_candidates:
				if cand.collide_point(*touch.pos):
					self.app.selection = cand
					cand.selected = True
					touch.grab(cand)
					ret = super().on_touch_move(touch)
					return ret

	def portal_touch_up(self, touch):
		"""Try to create a portal between the spots the user chose."""
		try:
			# If the touch ended upon a spot, and there isn't
			# already a portal between the origin and this
			# destination, create one.
			destspot = next(self.spots_at(*touch.pos))
			orig = self.origspot.proxy
			dest = destspot.proxy
			if not (orig.name in self.character.portal
					and dest.name in self.character.portal[orig.name]):
				symmetrical = hasattr(self, 'protoportal2') and not (
					orig.name in self.character.preportal
					and dest.name in self.character.preportal[orig.name])
				port = self.character.new_portal(orig.name,
													dest.name,
													symmetrical=symmetrical)
				self.arrowlayout.add_widget(self.make_arrow(port))
				if symmetrical:
					self.arrowlayout.add_widget(
						self.make_arrow(
							self.character.portal[dest.name][orig.name]))
		except StopIteration:
			pass
		self.remove_widget(self.protoportal)
		if hasattr(self, 'protoportal2'):
			self.remove_widget(self.protoportal2)
			del self.protoportal2
		self.remove_widget(self.protodest)
		del self.protoportal
		del self.protodest

	def on_touch_up(self, touch):
		"""Delegate touch handling if possible, else select something."""
		if hasattr(self, '_lasttouch') and self._lasttouch == touch:
			return
		self._lasttouch = touch
		touch.push()
		touch.apply_transform_2d(self.to_local)
		if hasattr(self, 'protodest'):
			Logger.debug("Board: on_touch_up making a portal")
			touch.ungrab(self)
			ret = self.portal_touch_up(touch)
			touch.pop()
			return ret
		if self.app.selection and hasattr(self.app.selection, 'on_touch_up'):
			self.app.selection.dispatch('on_touch_up', touch)
		for candidate in self.selection_candidates:
			if candidate.collide_point(*touch.pos):
				if hasattr(candidate, 'selected'):
					if isinstance(candidate,
									GraphArrow) and candidate.reciprocal:
						# mirror arrows can't be selected directly, you have to work with the one they mirror
						if candidate.portal.get('is_mirror', False):
							candidate.selected = True
							candidate = candidate.reciprocal
						elif candidate.reciprocal and candidate.reciprocal.portal.get(
							'is_mirror', False):
							candidate.reciprocal.selected = True
					candidate.selected = True
				if hasattr(self.app.selection, 'selected'):
					self.app.selection.selected = False
					if isinstance(self.app.selection,
									GraphArrow) and self.app.selection.reciprocal \
                                     and candidate is not self.app.selection.reciprocal:
						self.app.selection.reciprocal.selected = False
				self.app.selection = candidate
				self.keep_selection = True
				parent = candidate.parent
				parent.remove_widget(candidate)
				parent.add_widget(candidate)
				break
		if not self.keep_selection:
			Logger.debug("Board: deselecting " + repr(self.app.selection))
			if hasattr(self.app.selection, 'selected'):
				self.app.selection.selected = False
				if isinstance(self.app.selection,
								GraphArrow) and self.app.selection.reciprocal:
					self.app.selection.reciprocal.selected = False
			self.app.selection = None
		self.keep_selection = False
		touch.ungrab(self)
		touch.pop()
		return

	def _pull_size(self, *args):
		if self.wallpaper.texture is None:
			Clock.schedule_once(self._pull_size, 0.001)
			return
		self.size = self.wallpaper.size = self.wallpaper.texture.size

	def _pull_image(self, *args):
		self.wallpaper.source = self.wallpaper_path
		self._pull_size()

	def on_parent(self, *args):
		"""Create some subwidgets and trigger the first update."""
		if not self.parent or hasattr(self, '_parented'):
			return
		if not self.wallpaper_path:
			Logger.debug("Board: waiting for wallpaper_path")
			Clock.schedule_once(self.on_parent, 0)
			return
		self._parented = True
		self.wallpaper = Image(source=self.wallpaper_path)
		self.bind(wallpaper_path=self._pull_image)
		self._pull_size()
		self.kvlayoutback = KvLayoutBack(**self.widkwargs)
		self.arrowlayout = ArrowLayout(**self.widkwargs)
		self.spotlayout = FinalLayout(**self.widkwargs)
		self.kvlayoutfront = KvLayoutFront(**self.widkwargs)
		for wid in self.wids:
			self.add_widget(wid)
			wid.pos = 0, 0
			wid.size = self.size
			if wid is not self.wallpaper:
				self.bind(size=wid.setter('size'))
		self.update()

	def on_character(self, *args):
		if self.character is None:
			return
		if self.parent is None:
			Clock.schedule_once(self.on_character, 0)
			return

		self.engine = getattr(self.character, 'engine', None)
		self.wallpaper_path = self.character.stat.setdefault(
			'wallpaper', 'wallpape.jpg')
		if '_control' not in self.character.stat or 'wallpaper' not in \
                self.character.stat[
			'_control']:
			control = self.character.stat.setdefault('_control', {})
			control['wallpaper'] = 'textinput'
			self.character.stat['_control'] = control
		self.character.stat.connect(self._trigger_pull_wallpaper)
		self.trigger_update()

	def pull_wallpaper(self, *args):
		self.wallpaper_path = self.character.stat.setdefault(
			'wallpaper', 'wallpape.jpg')

	def _trigger_pull_wallpaper(self, *args, **kwargs):
		if kwargs['key'] != 'wallpaper':
			return
		if hasattr(self, '_scheduled_pull_wallpaper'):
			Clock.unschedule(self._scheduled_pull_wallpaper)
		self._scheduled_pull_wallpaper = Clock.schedule_once(
			self.pull_wallpaper, 0)

	@trigger
	def kv_updated(self, *args):
		self.unbind(wallpaper_path=self.kvlayoutback.setter('wallpaper_path'))
		for wid in self.wids:
			self.remove_widget(wid)
		self.kvlayoutback = KvLayoutBack(pos=(0, 0),
											wallpaper_path=self.wallpaper_path)
		self.bind(wallpaper_path=self.kvlayoutback.setter('wallpaper_path'))
		self.kvlayoutfront = KvLayoutFront(**self.widkwargs)
		self.size = self.kvlayoutback.size
		self.kvlayoutback.bind(size=self.setter('size'))
		for wid in self.wids:
			self.add_widget(wid)

	def make_pawn(self, thing):
		"""Make a :class:`Pawn` to represent a :class:`Thing`, store it, and
        return it.

        """
		if thing["name"] in self.pawn:
			raise KeyError("Already have a Pawn for this Thing")
		r = self.pawn_cls(board=self, proxy=thing)
		self.pawn[thing["name"]] = r
		return r

	def make_spot(self, place):
		"""Make a :class:`Spot` to represent a :class:`Place`, store it, and
        return it.

        """
		if place["name"] in self.spot:
			raise KeyError("Already have a Spot for this Place")
		r = self.spot_cls(board=self, place=place)
		self.spot[place["name"]] = r
		if '_x' in place and '_y' in place:
			r.pos = (self.width * place['_x'], self.height * place['_y'])
		return r

	def make_arrow(self, portal):
		"""Make an :class:`Arrow` to represent a :class:`Portal`, store it,
        and return it.

        """
		if (portal["origin"] not in self.spot
			or portal["destination"] not in self.spot):
			raise ValueError("An :class:`Arrow` should only be made after "
								"the :class:`Spot`s it connects")
		if (portal["origin"] in self.arrow
			and portal["destination"] in self.arrow[portal["origin"]]):
			raise KeyError("Already have an Arrow for this Portal")
		return self._core_make_arrow(portal, self.spot[portal['origin']],
										self.spot[portal['destination']],
										self.arrow)

	def _core_make_arrow(self,
							portal,
							origspot,
							destspot,
							arrowmap,
							points=None):
		if points is None:
			r = self.arrow_cls(
				board=self,
				portal=portal,
				origspot=origspot,
				destspot=destspot,
			)
			r._trigger_repoint()
		else:
			r = self.arrow_cls(board=self,
								portal=portal,
								origspot=origspot,
								destspot=destspot,
								points=points)
		orign = portal["origin"]
		if orign not in arrowmap:
			arrowmap[orign] = {}
		arrowmap[orign][portal["destination"]] = r
		return r

	def rm_arrows_to_and_from(self, name):
		origs = list(self.arrow.keys())
		if name in origs:
			origs.remove(name)
			for dest in list(self.arrow[name].keys()):
				self.rm_arrow(name, dest)
		for orig in origs:
			if name in self.arrow[orig]:
				self.rm_arrow(orig, name)

	def rm_pawn(self, name, *args):
		"""Remove the :class:`Pawn` by the given name."""
		if name not in self.pawn:
			raise KeyError("No Pawn named {}".format(name))
		# Currently there's no way to connect Pawns with Arrows but I
		# think there will be, so, insurance
		self.rm_arrows_to_and_from(name)
		pwn = self.pawn.pop(name)
		if pwn in self.selection_candidates:
			self.selection_candidates.remove(pwn)
		pwn.parent.remove_widget(pwn)

	def _trigger_rm_pawn(self, name):
		Clock.schedule_once(partial(self.rm_pawn, name), 0)

	def rm_spot(self, name, *args):
		"""Remove the :class:`Spot` by the given name."""
		if name not in self.spot:
			raise KeyError("No Spot named {}".format(name))
		spot = self.spot.pop(name)
		if spot in self.selection_candidates:
			self.selection_candidates.remove(spot)
		pawns_here = list(spot.children)
		self.rm_arrows_to_and_from(name)
		self.spotlayout.remove_widget(spot)
		spot.canvas.clear()
		for pawn in pawns_here:
			self.rm_pawn(pawn.name)
		if name in self._scheduled_rm_spot:
			del self._scheduled_rm_spot[name]

	def _trigger_rm_spot(self, name):
		part = partial(self.rm_spot, name)
		if name in self._scheduled_rm_spot:
			Clock.unschedule(self._scheduled_rm_spot[name])
		self._scheduled_rm_spot[name] = Clock.schedule_once(part, 0)

	def rm_arrow(self, orig, dest, *args):
		"""Remove the :class:`Arrow` that goes from ``orig`` to ``dest``."""
		if (orig not in self.arrow or dest not in self.arrow[orig]):
			raise KeyError("No Arrow from {} to {}".format(orig, dest))
		arr = self.arrow[orig].pop(dest)
		if arr in self.selection_candidates:
			self.selection_candidates.remove(arr)
		self.arrowlayout.remove_widget(arr)
		if (orig, dest) in self._scheduled_rm_arrow:
			del self._scheduled_rm_arrow[orig, dest]

	def _trigger_rm_arrow(self, orig, dest):
		part = partial(self.rm_arrow, orig, dest)
		if (orig, dest) in self._scheduled_rm_arrow:
			Clock.unschedule(self._scheduled_rm_arrow[orig, dest])
		self._scheduled_rm_arrow[orig, dest] = Clock.schedule_once(part, 0)

	def graph_layout(self, graph):
		from networkx.drawing.layout import spring_layout
		return normalize_layout(spring_layout(graph))

	def discard_pawn(self, thingn, *args):
		if thingn in self.pawn:
			self.rm_pawn(thingn)
		if thingn in self._scheduled_discard_pawn:
			del self._scheduled_discard_pawn[thingn]

	def _trigger_discard_pawn(self, thing):
		part = partial(self.discard_pawn, thing)
		if thing in self._scheduled_discard_pawn:
			Clock.unschedule(self._scheduled_discard_pawn[thing])
		self._scheduled_discard_pawn[thing] = Clock.schedule_once(part, 0)

	def remove_absent_pawns(self, *args):
		Logger.debug("Board: removing pawns absent from {}".format(
			self.character.name))
		for pawn_name in list(self.pawn.keys()):
			if pawn_name not in self.character.thing:
				self.rm_pawn(pawn_name)

	def discard_spot(self, placen, *args):
		if placen in self.spot:
			self.rm_spot(placen)

	def _trigger_discard_spot(self, place):
		Clock.schedule_once(partial(self.discard_spot, place), 0)

	def remove_absent_spots(self, *args):
		Logger.debug("Board: removing spots absent from {}".format(
			self.character.name))
		for spot_name in list(self.spot.keys()):
			if spot_name not in self.character.place:
				self.rm_spot(spot_name)

	def discard_arrow(self, orign, destn, *args):
		if (orign in self.arrow and destn in self.arrow[orign]):
			self.rm_arrow(orign, destn)

	def _trigger_discard_arrow(self, orig, dest):
		Clock.schedule_once(partial(self.discard_arrow, orig, dest), 0)

	def remove_absent_arrows(self, *args):
		Logger.debug("Board: removing arrows absent from {}".format(
			self.character.name))
		for arrow_origin in list(self.arrow.keys()):
			for arrow_destination in list(self.arrow[arrow_origin].keys()):
				if (arrow_origin not in self.character.portal
					or arrow_destination
					not in self.character.portal[arrow_origin]):
					self.rm_arrow(arrow_origin, arrow_destination)

	def add_spot(self, placen, *args):
		if (placen in self.character.place and placen not in self.spot):
			self.spotlayout.add_widget(
				self.make_spot(self.character.place[placen]))

	def _trigger_add_spot(self, placen):
		Clock.schedule_once(partial(self.add_spot, placen), 0)

	def add_new_spots(self, *args):
		Logger.debug("Board: adding new spots to {}".format(
			self.character.name))
		places2add = []
		spots_unposd = []
		nodes_patch = {}
		placemap = self.character.place
		spotmap = self.spot
		default_image_paths = GraphSpot.default_image_paths
		for place_name in placemap:
			if place_name not in spotmap:
				place = placemap[place_name]
				places2add.append(place)
				patch = {}
				if '_image_paths' in place:
					zeroes = [0] * len(place['_image_paths'])
				else:
					patch['_image_paths'] = default_image_paths
					zeroes = [0]
				if '_offxs' not in place:
					patch['_offxs'] = zeroes
				if '_offys' not in place:
					patch['_offys'] = zeroes
				if patch:
					nodes_patch[place_name] = patch
		if nodes_patch:
			self.character.node.patch(nodes_patch)
		make_spot = self.make_spot
		spotlayout = self.spotlayout
		add_widget_to_spotlayout = spotlayout.add_widget
		for place in places2add:
			spot = make_spot(place)
			add_widget_to_spotlayout(spot)
			if '_x' not in place or '_y' not in place:
				spots_unposd.append(spot)
		self.spots_unposd = spots_unposd

	def add_arrow(self, orign, destn, *args):
		if not (orign in self.character.portal
				and destn in self.character.portal[orign]):
			raise ValueError("No portal for arrow {}->{}".format(orign, destn))
		if not (orign in self.arrow and destn in self.arrow[orign]):
			self.arrowlayout.add_widget(
				self.make_arrow(self.character.portal[orign][destn]))
		assert self.arrow[orign][destn] in self.arrowlayout.children

	def add_new_arrows(self, *args):
		Logger.debug("Board: adding new arrows to {}".format(
			self.character.name))
		portmap = self.character.portal
		arrowmap = self.arrow
		spotmap = self.spot
		arrowlayout = self.arrowlayout
		add_widget_to_arrowlayout = arrowlayout.add_widget
		core_make_arrow = self._core_make_arrow
		todo = []
		for arrow_orig, arrow_dests in portmap.items():
			for arrow_dest, portal in arrow_dests.items():
				if (arrow_orig not in arrowmap
					or arrow_dest not in arrowmap[arrow_orig]):
					todo.append(
						(portal, spotmap[arrow_orig], spotmap[arrow_dest]))
		points = get_points_multi(
			(origspot, destspot, 10) for (portal, origspot, destspot) in todo)
		for portal, origspot, destspot in todo:
			add_widget_to_arrowlayout(
				core_make_arrow(portal, origspot, destspot, arrowmap,
								points[origspot, destspot]))

	def add_pawn(self, thingn, *args):
		if thingn not in self.character.thing:
			raise KeyError(f"No Thing for pawn: {thingn}")
		if thingn in self.pawn:
			raise KeyError(f"Already have pawn for Thing: {thingn}")
		pwn = self.make_pawn(self.character.thing[thingn])
		whereat = self.spot[pwn.loc_name]
		whereat.add_widget(pwn)
		self.pawn[thingn] = pwn
		if thingn in self._scheduled_add_pawn:
			del self._scheduled_add_pawn[thingn]

	def _trigger_add_pawn(self, thingn):
		part = partial(self.add_pawn, thingn)
		if thingn in self._scheduled_add_pawn:
			Clock.unschedule(self._scheduled_add_pawn[thingn])
		self._scheduled_add_pawn[thingn] = Clock.schedule_once(part, 0)

	def add_new_pawns(self, *args):
		Logger.debug("Board: adding new pawns to {}".format(
			self.character.name))
		nodes_patch = {}
		things2add = []
		pawns_added = []
		pawnmap = self.pawn
		for (thing_name, thing) in self.character.thing.items():
			if thing_name not in pawnmap:
				things2add.append(thing)
				patch = {}
				if '_image_paths' in thing:
					zeroes = [0] * len(thing['_image_paths'])
				else:
					patch['_image_paths'] = Pawn.default_image_paths
					zeroes = [0] * len(Pawn.default_image_paths)
				if '_offxs' not in thing:
					patch['_offxs'] = zeroes
				if '_offys' not in thing:
					patch['_offys'] = zeroes
				if '_stackhs' not in thing:
					patch['_stackhs'] = zeroes
				if patch:
					nodes_patch[thing_name] = patch
		if nodes_patch:
			self.character.node.patch(nodes_patch)
		make_pawn = self.make_pawn
		spotmap = self.spot
		for thing in things2add:
			pwn = make_pawn(thing)
			pawns_added.append(pwn)
			whereat = spotmap[thing['location']]
			whereat.add_widget(pwn)

	def update(self, *args):
		"""Force an update to match the current state of my character.

        This polls every element of the character, and therefore
        causes me to sync with the LiSE core for a long time. Avoid
        when possible.

        """
		if not hasattr(self, 'engine') or getattr(self.engine, 'closed',
													False):
			Logger.warning(
				"Board: tried to update without a connection to a LiSE core")
			return
		if not self.spotlayout or not self.arrowlayout:
			self.trigger_update()
			return
		# remove widgets that don't represent anything anymore
		Logger.debug("Board: updating")
		self.remove_absent_pawns()
		self.remove_absent_spots()
		self.remove_absent_arrows()
		# add widgets to represent new stuff
		self.add_new_spots()
		if self.arrow_cls:
			self.add_new_arrows()
		self.add_new_pawns()
		self.spotlayout.finalize_all()
		Logger.debug("Board: updated")

	trigger_update = trigger(update)

	def update_from_delta(self, delta, *args):
		"""Apply the changes described in the dict ``delta``."""
		for (node, extant) in delta.get('nodes', {}).items():
			if extant:
				if node in delta.get('node_val', {}) \
                              and 'location' in delta['node_val'][node] \
                              and node not in self.pawn:
					self.add_pawn(node)
				elif node not in self.spot:
					self.add_spot(node)
					spot = self.spot[node]
					if '_x' not in spot.place or '_y' not in spot.place:
						self.spots_unposd.append(spot)
			else:
				if node in self.pawn:
					self.rm_pawn(node)
				if node in self.spot:
					self.rm_spot(node)
		for (node, stats) in delta.get('node_val', {}).items():
			if node in self.spot:
				spot = self.spot[node]
				x = stats.get('_x')
				y = stats.get('_y')
				if x is not None:
					spot.x = x * self.width
				if y is not None:
					spot.y = y * self.height
				if '_image_paths' in stats:
					spot.paths = stats[
						'_image_paths'] or spot.default_image_paths
			elif node in self.pawn:
				pawn = self.pawn[node]
				if 'location' in stats:
					pawn.loc_name = stats['location']
				if '_image_paths' in stats:
					pawn.paths = stats[
						'_image_paths'] or pawn.default_image_paths
			else:
				Logger.warning(
					"Board: diff tried to change stats of node {} "
					"but I don't have a widget for it".format(node))
		for (orig, dests) in delta.get('edges', {}).items():
			for (dest, extant) in dests.items():
				if extant and (orig not in self.arrow
								or dest not in self.arrow[orig]):
					self.add_arrow(orig, dest)
				elif not extant and orig in self.arrow and dest in self.arrow[
					orig]:
					self.rm_arrow(orig, dest)

	def trigger_update_from_delta(self, delta, *args):
		part = partial(self.update_from_delta, delta)
		if hasattr(self, '_scheduled_update_from_delta'):
			Clock.unschedule(self._scheduled_update_from_delta)
		self._scheduled_update_from_delta = Clock.schedule_once(part, 0)

	def on_spots_unposd(self, *args):
		# TODO: If only some spots are unpositioned, and they remain
		# that way for several frames, put them somewhere that the
		# user will be able to find.
		if not self.spots_unposd:
			return
		try:
			self.grid_layout()
		except (TypeError, ValueError):
			self.nx_layout()

	def _apply_node_layout(self, l, *args):
		if self.width == 1 or self.height == 1:
			Clock.schedule_once(partial(self._apply_node_layout, l), 0.01)
			return
		node_upd = {}
		for spot in self.spots_unposd:
			(x, y) = l[spot.name]
			assert 0 <= x <= 0.99, "{} has invalid x: {}".format(spot.name, x)
			assert 0 <= y <= 0.99, "{} has invalid y: {}".format(spot.name, y)
			assert spot in self.spotlayout.children
			assert self.spotlayout.width == self.width
			assert self.spotlayout.height == self.height
			node_upd[spot.name] = {'_x': x, '_y': y}
			spot.pos = (x * self.width, y * self.height)
		if node_upd:
			self.character.node.patch(node_upd)
		self.spots_unposd = []

	def grid_layout(self, *args):
		self._apply_node_layout(
			normalize_layout(
				{spot.name: spot.name
					for spot in self.spots_unposd}))

	def nx_layout(self, *args):
		for spot in self.spots_unposd:
			if not (spot.name and spot.proxy):
				Clock.schedule_once(self.nx_layout, 0)
				return
		spots_only = self.character.facade()
		for thing in list(spots_only.thing.keys()):
			del spots_only.thing[thing]
		self._apply_node_layout(self.graph_layout(spots_only))

	def arrows(self):
		"""Iterate over all my arrows."""
		for o in self.arrow.values():
			for arro in o.values():
				yield arro

	def pawns_at(self, x, y):
		"""Iterate over pawns that collide the given point."""
		for pawn in self.pawn.values():
			if pawn.collide_point(x, y):
				yield pawn

	def spots_at(self, x, y):
		"""Iterate over spots that collide the given point."""
		for spot in self.spot.values():
			if spot.collide_point(x, y):
				yield spot

	def arrows_at(self, x, y):
		"""Iterate over arrows that collide the given point."""
		for arrow in self.arrows():
			if arrow.collide_point(x, y):
				yield arrow


class BoardScatterPlane(ScatterPlane):
	selection_candidates = ListProperty([])
	selection = ObjectProperty(allownone=True)
	keep_selection = BooleanProperty(False)
	board = ObjectProperty()
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty()

	def spot_from_dummy(self, dummy):
		"""Make a real place and its spot from a dummy spot.

        Create a new :class:`graph.Spot` instance, along with the
        underlying :class:`LiSE.Place` instance, and give it the name,
        position, and imagery of the provided dummy.

        """
		(x, y) = self.to_local(*dummy.pos_up)
		x /= self.board.width
		y /= self.board.height
		self.board.spotlayout.add_widget(
			self.board.make_spot(
				self.board.character.new_place(dummy.name,
												_x=x,
												_y=y,
												_image_paths=list(
													dummy.paths))))
		dummy.num += 1

	def pawn_from_dummy(self, dummy):
		"""Make a real thing and its pawn from a dummy pawn.

        Create a new :class:`graph.Pawn` instance, along with the
        underlying :class:`LiSE.Thing` instance, and give it the name,
        location, and imagery of the provided dummy.

        """
		candidates = []
		dummy_center = self.to_local(*dummy.center)
		dummy.pos = self.to_local(*dummy.pos)
		for spot in self.board.spot.values():
			if spot.collide_widget(dummy):
				candidates.append(spot)
		if not candidates:
			return
		whereat = candidates.pop()
		if candidates:
			dist = Vector(*whereat.center).distance(dummy_center)
			while candidates:
				thereat = candidates.pop()
				thereto = Vector(*thereat.center).distance(dummy_center)
				if thereto < dist:
					whereat, dist = thereat, thereto
		whereat.add_widget(
			self.board.make_pawn(
				self.board.character.new_thing(dummy.name,
												whereat.proxy.name,
												_image_paths=list(
													dummy.paths))))
		dummy.num += 1

	def on_board(self, *args):
		if hasattr(self, '_oldboard'):
			self.unbind(
				adding_portal=self._oldboard.setter('adding_portal'),
				reciprocal_portal=self._oldboard.setter('reciprocal_portal'))
		self.clear_widgets()
		self.add_widget(self.board)
		self.board.adding_portal = self.adding_portal
		self.board.reciprocal_portal = self.reciprocal_portal
		self.bind(adding_portal=self.board.setter('adding_portal'),
					reciprocal_portal=self.board.setter('reciprocal_portal'))
		self._oldboard = self.board

	def on_touch_down(self, touch):
		if touch.is_mouse_scrolling:
			scale = self.scale + (0.05
									if touch.button == 'scrolldown' else -0.05)
			if (self.scale_min and scale < self.scale_min) \
                       or (self.scale_max and scale > self.scale_max):
				return
			rescale = scale * 1.0 / self.scale
			self.apply_transform(Matrix().scale(rescale, rescale, rescale),
									post_multiply=True,
									anchor=self.to_local(*touch.pos))
			return self.dispatch('on_transform_with_touch', touch)
		return super().on_touch_down(touch)

	def apply_transform(self, trans, post_multiply=False, anchor=(0, 0)):
		super().apply_transform(trans,
								post_multiply=post_multiply,
								anchor=anchor)
		self._last_transform = trans, post_multiply, anchor

	def on_transform_with_touch(self, touch):
		x, y = self.pos
		w = self.board.width * self.scale
		h = self.board.height * self.scale
		if hasattr(self, '_last_transform') and (w < self.parent.width
													or h < self.parent.height):
			trans, post_multiply, anchor = self._last_transform
			super().apply_transform(trans.inverse(), post_multiply, anchor)
			return
		if x > self.parent.x:
			self.x = self.parent.x
		if y > self.parent.y:
			self.y = self.parent.y
		if x + w < self.parent.right:
			self.x = self.parent.right - w
		if y + h < self.parent.top:
			self.y = self.parent.top - h


class GraphBoardView(BoardView):
	adding_portal = BooleanProperty(False)
	reciprocal_portal = BooleanProperty(True)


Builder.load_string("""
<GraphBoard>:
    app: app
    size_hint: None, None
<GraphBoardView>:
    plane: boardplane
    BoardScatterPlane:
        id: boardplane
        board: root.board
        adding_portal: root.adding_portal
        reciprocal_portal: root.reciprocal_portal
        scale_min: root.scale_min
        scale_max: root.scale_max
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock
from kivy.properties import NumericProperty, ObjectProperty
from kivy.uix.stencilview import StencilView


class BoardView(StencilView):
	board = ObjectProperty()
	plane = ObjectProperty()
	scale_min = NumericProperty(allownone=True)
	scale_max = NumericProperty(allownone=True)

	def on_pos(self, *args):
		if self.board and self.children:
			self.children[0].pos = self.pos
		else:
			Clock.schedule_once(self.on_pos, 0.001)

	def on_size(self, *args):
		if self.board and self.children:
			self.children[0].size = self.size
		else:
			Clock.schedule_once(self.on_size, 0.001)

	def on_parent(self, *args):
		if self.board and self.children:
			self.children[0].pos = self.pos
			self.children[0].size = self.size
		else:
			Clock.schedule_once(self.on_parent, 0.001)

	def spot_from_dummy(self, dummy):
		self.plane.spot_from_dummy(dummy)

	def pawn_from_dummy(self, dummy):
		self.plane.pawn_from_dummy(dummy)



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.app import App
from kivy.clock import Clock
from kivy.properties import (DictProperty, NumericProperty, StringProperty,
								ObjectProperty)
from kivy.lang import Builder
from kivy.uix.screenmanager import Screen
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.textinput import TextInput
from .statlist import BaseStatListView


class FloatInput(TextInput):

	def insert_text(self, s, from_undo=False):
		return super().insert_text(''.join(c for c in s if c in '0123456789.'),
									from_undo)


class ControlTypePicker(Button):
	app = ObjectProperty()
	key = ObjectProperty()
	mainbutton = ObjectProperty()
	dropdown = ObjectProperty()
	set_control = ObjectProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.build()

	def set_value(self, k, v):
		if v is None:
			del self.app.selected_proxy[k]
		else:
			self.app.selected_proxy[k] = v

	def build(self, *args):
		if None in (self.key, self.set_control):
			Clock.schedule_once(self.build, 0)
			return
		self.mainbutton = None
		self.dropdown = None
		self.dropdown = DropDown()
		self.dropdown.bind(
			on_select=lambda instance, x: self.set_control(self.key, x))
		readoutbut = Button(text='readout',
							size_hint_y=None,
							height=self.height,
							background_color=(0.7, 0.7, 0.7, 1))
		readoutbut.bind(
			on_release=lambda instance: self.dropdown.select('readout'))
		self.dropdown.add_widget(readoutbut)
		textinbut = Button(text='textinput',
							size_hint_y=None,
							height=self.height,
							background_color=(0.7, 0.7, 0.7, 1))
		textinbut.bind(
			on_release=lambda instance: self.dropdown.select('textinput'))
		self.dropdown.add_widget(textinbut)
		togbut = Button(text='togglebutton',
						size_hint_y=None,
						height=self.height,
						background_color=(0.7, 0.7, 0.7, 1))
		togbut.bind(
			on_release=lambda instance: self.dropdown.select('togglebutton'))
		self.dropdown.add_widget(togbut)
		sliderbut = Button(text='slider',
							size_hint_y=None,
							height=self.height,
							background_color=(0.7, 0.7, 0.7, 1))
		sliderbut.bind(
			on_release=lambda instance: self.dropdown.select('slider'))
		self.dropdown.add_widget(sliderbut)
		self.bind(on_release=self.dropdown.open)


class ConfigListItemToggleButton(BoxLayout):
	true_text = StringProperty('0')
	false_text = StringProperty('1')

	def set_true_text(self, *args):
		self.parent.set_config(self.parent.key, 'true_text',
								self.ids.truetext.text)
		self.true_text = self.ids.truetext.text

	def set_false_text(self, *args):
		self.parent.set_config(self.parent.key, 'false_text',
								self.ids.falsetext.text)


class ConfigListItemSlider(BoxLayout):
	min = NumericProperty(0.)
	max = NumericProperty(1.)

	def set_min(self, *args):
		minn = float(self.ids.minimum.text)
		try:
			self.parent.set_config(self.parent.key, 'min', minn)
			self.min = minn
		except ValueError:
			self.ids.minimum.text = ''

	def set_max(self, *args):
		maxx = float(self.ids.minimum.text)
		try:
			self.parent.set_config(self.parent.key, 'max', maxx)
			self.max = maxx
		except ValueError:
			self.ids.maximum.text = ''


class ConfigListItemCustomizer(BoxLayout):
	key = ObjectProperty()
	control = StringProperty()
	config = DictProperty()
	set_config = ObjectProperty()

	def on_control(self, *args):
		self.clear_widgets()
		if self.control == 'togglebutton':
			if 'true_text' not in self.config or 'false_text' not in self.config:
				Clock.schedule_once(self.on_control, 0)
				return
			wid = ConfigListItemToggleButton(
				true_text=self.config['true_text'],
				false_text=self.config['false_text'])
			self.add_widget(wid)
		elif self.control == 'slider':
			if 'min' not in self.config or 'max' not in self.config:
				Clock.schedule_once(self.on_control, 0)
				return
			wid = ConfigListItemSlider(min=self.config['min'],
										max=self.config['max'])
			self.add_widget(wid)


class ConfigListItem(BoxLayout):
	key = ObjectProperty()
	config = DictProperty()
	set_control = ObjectProperty()
	set_config = ObjectProperty()
	deleter = ObjectProperty()


class StatListViewConfigurator(BaseStatListView):
	statlist = ObjectProperty()
	_key_cfg_setters = DictProperty()
	_val_text_setters = DictProperty()
	_control_wids = DictProperty()

	def set_control(self, key, value):
		config = self.proxy.get('_config', {})
		if value == 'slider':
			if 'min' not in config:
				self.set_config(key, 'min', 0.0)
			if 'max' not in config:
				self.set_config(key, 'max', 1.0)
		elif value == 'togglebutton':
			if 'true_text' not in config:
				self.set_config(key, 'true_text', '1')
			if 'false_text' not in config:
				self.set_config(key, 'false_text', '0')
		self.set_config(key, 'control', value)

	def munge(self, k, v):
		# makes ConfigListItem
		ret = super().munge(k, v)
		ret['deleter'] = self.del_key
		ret['set_control'] = self.set_control
		ret['set_config'] = self.set_config
		return ret


class StatScreen(Screen):
	statlist = ObjectProperty()
	statcfg = ObjectProperty()
	toggle = ObjectProperty()
	proxy = ObjectProperty()

	@property
	def engine(self):
		return App.get_running_app().engine

	def new_stat(self):
		"""Look at the key and value that the user has entered into the stat
        configurator, and set them on the currently selected
        entity.

        """
		key = self.ids.newstatkey.text
		value = self.ids.newstatval.text
		if not (key and value):
			# TODO implement some feedback to the effect that
			# you need to enter things
			return
		try:
			self.proxy[key] = self.engine.unpack(value)
		except (TypeError, ValueError):
			self.proxy[key] = value
		self.ids.newstatkey.text = ''
		self.ids.newstatval.text = ''


Builder.load_string("""
<ConfigListItemCustomization>:
    pos_hint: {'x': 0, 'y': 0}
<ConfigListItemToggleButton>:
    Label:
        text: 'True text:'
    TextInput:
        id: truetext
        hint_text: root.true_text
        on_text_validate: root.set_true_text()
    Label:
        text: 'False text:'
    TextInput:
        id: falsetext
        hint_text: root.false_text
        on_text_validate: root.set_false_text()
<ConfigListItemSlider>:
    Label:
        text: 'Minimum:'
    TextInput:
        id: minimum
        hint_text: str(root.min)
        on_text_validate: root.set_min()
    Label:
        text: 'Maximum:'
    TextInput:
        id: maximum
        hint_text: str(root.max)
        on_text_validate: root.set_max()
<ConfigListItem>:
    height: 30
    Button:
        size_hint_x: 0.4 / 3
        text: 'del'
        on_release: root.deleter(root.key)
    Label:
        size_hint_x: 0.4 / 3
        text: str(root.key)
    ControlTypePicker:
        size_hint_x: 0.4 / 3
        key: root.key
        set_control: root.set_control
        text: root.config['control'] if 'control' in root.config else 'readout'
    ConfigListItemCustomizer:
        size_hint_x: 0.6
        control: root.config['control'] if 'control' in root.config else 'readout'
        config: root.config
        key: root.key
        set_config: root.set_config
<StatScreen>:
    name: 'statcfg'
    statcfg: cfg
    BoxLayout:
        orientation: 'vertical'
        StatListViewConfigurator:
            viewclass: 'ConfigListItem'
            id: cfg
            app: app
            engine: root.engine
            proxy: root.proxy
            statlist: root.statlist
            size_hint_y: 0.95
            RecycleBoxLayout:
                default_size: None, dp(56)
                default_size_hint: 1, None
                size_hint_y: None
                height: self.minimum_height
                orientation: 'vertical'
        BoxLayout:
            orientation: 'horizontal'
            size_hint_y: 0.05
            TextInput:
                id: newstatkey
                multiline: False
                write_tab: False
                hint_text: 'New stat'
            TextInput:
                id: newstatval
                multiline: False
                write_tab: False
                hint_text: 'Value'
            Button:
                id: newstatbut
                text: '+'
                on_release: root.new_stat()
            Button:
                id: closer
                text: 'Close'
                on_release: root.toggle()
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock
from kivy.properties import ObjectProperty

from ELiDE.kivygarden.texturestack import ImageStack
from .util import trigger


class ImageStackProxy(ImageStack):
	proxy = ObjectProperty()
	name = ObjectProperty()

	def finalize(self, initial=True):
		if getattr(self, '_finalized', False):
			return
		if (self.proxy is None or not hasattr(self.proxy, 'name')):
			Clock.schedule_once(self.finalize, 0)
			return
		if initial:
			self.name = self.proxy.name
			if '_image_paths' in self.proxy:
				try:
					self.paths = self.proxy['_image_paths']
				except Exception as ex:
					if not ex.args[0].startswith('Unable to load image type'):
						raise ex
					self.paths = self.default_image_paths
			else:
				self.paths = self.proxy.setdefault('_image_paths',
													self.default_image_paths)
		self.bind(paths=self._trigger_push_image_paths,
					offxs=self._trigger_push_offxs,
					offys=self._trigger_push_offys)
		self._finalized = True
		self.finalize_children()

	def finalize_children(self, *args):
		for child in self.children:
			child.finalize()
		self.bind(children=self._trigger_finalize_children)

	_trigger_finalize_children = trigger(finalize_children)

	def unfinalize(self):
		self.unbind(paths=self._trigger_push_image_paths,
					offxs=self._trigger_push_offxs,
					offys=self._trigger_push_offys)
		self._finalized = False

	def pull_from_proxy(self, *args):
		initial = not hasattr(self, '_finalized')
		self.unfinalize()
		for key, att in [('_image_paths', 'paths'), ('_offxs', 'offxs'),
							('_offys', 'offys')]:
			if key in self.proxy and self.proxy[key] != getattr(self, att):
				setattr(self, att, self.proxy[key])
		self.finalize(initial)

	def _trigger_pull_from_proxy(self, *args, **kwargs):
		if hasattr(self, '_scheduled_pull_from_proxy'):
			Clock.unschedule(self._scheduled_pull_from_proxy)
		self._scheduled_pull_from_proxy = Clock.schedule_once(
			self.pull_from_proxy, 0)

	@trigger
	def _trigger_push_image_paths(self, *args):
		self.proxy['_image_paths'] = list(self.paths)

	@trigger
	def _trigger_push_offxs(self, *args):
		self.proxy['_offxs'] = list(self.offxs)

	@trigger
	def _trigger_push_offys(self, *args):
		self.proxy['_offys'] = list(self.offys)

	@trigger
	def _trigger_push_stackhs(self, *args):
		self.proxy['_stackhs'] = list(self.stackhs)

	@trigger
	def restack(self, *args):
		childs = sorted(list(self.children),
						key=lambda child: child.priority,
						reverse=True)
		self.clear_widgets()
		for child in childs:
			self.add_widget(child)
		self.do_layout()



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Code that draws the box around a Pawn or Spot when it's selected"""
from collections import defaultdict
from functools import partial

from kivy.properties import ObjectProperty, BooleanProperty, ListProperty
from kivy.graphics import (InstructionGroup, Translate, PopMatrix, PushMatrix,
							Color, Line)
from kivy.uix.layout import Layout
from kivy.clock import Clock
from ..util import trigger
from ..imagestackproxy import ImageStackProxy


class GraphPawnSpot(ImageStackProxy, Layout):
	"""The kind of ImageStack that represents a :class:`Thing` or
    :class:`Place`.

    """
	board = ObjectProperty()
	engine = ObjectProperty()
	selected = BooleanProperty(False)
	linecolor = ListProperty()
	selected_outline_color = ListProperty([0, 1, 1, 1])
	unselected_outline_color = ListProperty([0, 0, 0, 0])
	use_boardspace = True

	def __init__(self, **kwargs):
		if 'proxy' in kwargs:
			kwargs['name'] = kwargs['proxy'].name
		super().__init__(**kwargs)
		self.bind(pos=self._position)

	def on_touch_move(self, touch):
		"""If I'm being dragged, move to follow the touch."""
		if touch.grab_current is not self:
			return False
		self.center = touch.pos
		return True

	def finalize(self, initial=True):
		"""Call this after you've created all the PawnSpot you need and are ready to add them to the board."""
		if getattr(self, '_finalized', False):
			return
		if (self.proxy is None or not hasattr(self.proxy, 'name')):
			Clock.schedule_once(partial(self.finalize, initial=initial), 0)
			return
		if initial:
			self.name = self.proxy.name
			if '_image_paths' in self.proxy:
				try:
					self.paths = self.proxy['_image_paths']
				except Exception as ex:
					if not ex.args[0].startswith('Unable to load image type'):
						raise ex
					self.paths = self.default_image_paths
			else:
				self.paths = self.proxy.setdefault('_image_paths',
													self.default_image_paths)
			zeroes = [0] * len(self.paths)
			self.offxs = self.proxy.setdefault('_offxs', zeroes)
			self.offys = self.proxy.setdefault('_offys', zeroes)
			self.proxy.connect(self._trigger_pull_from_proxy)
		self.bind(paths=self._trigger_push_image_paths,
					offxs=self._trigger_push_offxs,
					offys=self._trigger_push_offys)
		self._finalized = True
		self.finalize_children()

	def unfinalize(self):
		self.unbind(paths=self._trigger_push_image_paths,
					offxs=self._trigger_push_offxs,
					offys=self._trigger_push_offys)
		self._finalized = False

	def pull_from_proxy(self, *args):
		initial = not hasattr(self, '_finalized')
		self.unfinalize()
		for key, att in [('_image_paths', 'paths'), ('_offxs', 'offxs'),
							('_offys', 'offys')]:
			if key in self.proxy and self.proxy[key] != getattr(self, att):
				setattr(self, att, self.proxy[key])
		self.finalize(initial)

	def _trigger_pull_from_proxy(self, *args, **kwargs):
		Clock.unschedule(self.pull_from_proxy)
		Clock.schedule_once(self.pull_from_proxy, 0)

	@trigger
	def _trigger_push_image_paths(self, *args):
		self.proxy['_image_paths'] = list(self.paths)

	@trigger
	def _trigger_push_offxs(self, *args):
		self.proxy['_offxs'] = list(self.offxs)

	@trigger
	def _trigger_push_offys(self, *args):
		self.proxy['_offys'] = list(self.offys)

	@trigger
	def _trigger_push_stackhs(self, *args):
		self.proxy['_stackhs'] = list(self.stackhs)

	def on_linecolor(self, *args):
		"""If I don't yet have the instructions for drawing the selection box
        in my canvas, put them there. In any case, set the
        :class:`Color` instruction to match my current ``linecolor``.

        """
		if hasattr(self, 'color'):
			self.color.rgba = self.linecolor
			return

		def upd_box_translate(*args):
			self.box_translate.xy = self.pos

		def upd_box_points(*args):
			self.box.points = [
				0, 0, self.width, 0, self.width, self.height, 0, self.height,
				0, 0
			]

		self.boxgrp = boxgrp = InstructionGroup()
		self.color = Color(*self.linecolor)
		self.box_translate = Translate(*self.pos)
		boxgrp.add(PushMatrix())
		boxgrp.add(self.box_translate)
		boxgrp.add(self.color)
		self.box = Line()
		upd_box_points()
		self.bind(size=upd_box_points, pos=upd_box_translate)
		boxgrp.add(self.box)
		boxgrp.add(Color(1., 1., 1.))
		boxgrp.add(PopMatrix())

	def on_board(self, *args):
		if not (hasattr(self, 'group') and hasattr(self, 'boxgrp')):
			Clock.schedule_once(self.on_board, 0)
			return
		self.canvas.add(self.group)
		self.canvas.add(self.boxgrp)

	def add_widget(self, wid, index=None, canvas=None):
		if index is None:
			for index, child in enumerate(self.children, start=1):
				if wid.priority < child.priority:
					index = len(self.children) - index
					break
		super().add_widget(wid, index=index, canvas=canvas)
		self._trigger_layout()

	def remove_widget(self, widget):
		super().remove_widget(widget)
		self._trigger_layout()

	def do_layout(self, *args):
		# First try to lay out my children inside of me,
		# leaving at least this much space on the sides
		xpad = self.proxy.get('_xpad', self.width / 4)
		ypad = self.proxy.get('_ypad', self.height / 4)
		self.gutter = gutter = self.proxy.get('_gutter', xpad / 2)
		height = self.height - ypad
		content_height = 0
		too_tall = False
		width = self.width - xpad
		content_width = 0
		groups = defaultdict(list)
		for child in self.children:
			group = child.proxy.get('_group', '')
			groups[group].append(child)
			if child.height > height:
				height = child.height
				too_tall = True
		piles = {}
		# Arrange the groups into piles that will fit in me vertically
		for group, members in groups.items():
			members.sort(key=lambda x: x.width * x.height, reverse=True)
			high = 0
			subgroups = []
			subgroup = []
			for member in members:
				high += member.height
				if high > height:
					subgroups.append(subgroup)
					subgroup = [member]
					high = member.height
				else:
					subgroup.append(member)
			subgroups.append(subgroup)
			content_height = max(
				(content_height, sum(wid.height for wid in subgroups[0])))
			content_width += sum(
				max(wid.width for wid in subgrp) for subgrp in subgroups)
			piles[group] = subgroups
		self.content_width = content_width + gutter * (len(piles) - 1)
		too_wide = content_width > width
		# If I'm big enough to fit all this stuff, calculate an offset that will ensure
		# it's all centered. Otherwise just offset to my top-right so the user can still
		# reach me underneath all the pawns.
		if too_wide:
			offx = self.width
		else:
			offx = self.width / 2 - content_width / 2
		if too_tall:
			offy = self.height
		else:
			offy = self.height / 2 - content_height / 2
		for pile, subgroups in sorted(piles.items()):
			for subgroup in subgroups:
				subw = subh = 0
				for member in subgroup:
					rel_y = offy + subh
					member.rel_pos = (offx, rel_y)
					x, y = self.pos
					member.pos = x + offx, y + rel_y
					subw = max((subw, member.width))
					subh += member.height
				offx += subw
			offx += gutter

	def _position(self, *args):
		x, y = self.pos
		for child in self.children:
			offx, offy = getattr(child, 'rel_pos', (0, 0))
			child.pos = x + offx, y + offy

	def on_selected(self, *args):
		if self.selected:
			self.linecolor = self.selected_outline_color
		else:
			self.linecolor = self.unselected_outline_color



from kivy.base import EventLoop
from kivy.tests import UnitTestTouch

from .util import ELiDEAppTest, idle_until, window_with_widget
from LiSE import Engine


class StringsEditorTest(ELiDEAppTest):

	def test_strings_editor(self):
		assert 'LiSE' in self.app.config
		app = self.app
		print('app', id(app))
		win = window_with_widget(app.build())
		idle_until(lambda: hasattr(app, 'mainscreen'), 100,
					"app never got mainscreen")
		idle_until(lambda: app.manager.has_screen('timestream'), 100,
					"timestream never added to manager")
		idle_until(lambda: hasattr(app.mainmenu, 'configurator'), 100,
					'DirPicker never got configurator')
		app.mainmenu.configurator.start()  # start with blank world
		idle_until(lambda: app.engine, 100, "app never got engine")
		idle_until(lambda: app.strings.children, 100,
					"strings never got children")
		idle_until(lambda: app.strings.edbox, 100, 'strings never got edbox')
		idle_until(lambda: 'physical' in app.mainscreen.graphboards, 100,
					'never got physical in graphboards')
		edbox = app.strings.edbox
		strings_list = edbox.ids.strings_list
		idle_until(lambda: strings_list.store, 100,
					"strings_list never got store")
		strings_ed = edbox.ids.strings_ed
		app.strings.toggle()
		idle_until(lambda: strings_list.data, 100,
					"strings_list never got data")
		self.advance_frames(10)
		touchy = UnitTestTouch(*strings_ed.ids.stringname.center)
		touchy.touch_down()
		EventLoop.idle()
		touchy.touch_up()
		EventLoop.idle()
		strings_ed.ids.stringname.text = 'a string'
		idle_until(lambda: strings_ed.name == 'a string', 100,
					"name never set")
		touchier = UnitTestTouch(*strings_ed.ids.string.center)
		touchier.touch_down()
		EventLoop.idle()
		touchier.touch_up()
		self.advance_frames(10)
		strings_ed.ids.string.text = 'its value'
		idle_until(lambda: strings_ed.source == 'its value', 100,
					'source never set')
		self.advance_frames(10)
		edbox.dismiss()
		app.stop()
		with Engine(self.prefix) as eng:
			assert 'a string' in eng.string
			assert eng.string['a string'] == 'its value'



from kivy.tests.common import UnitTestTouch
from kivy.tests.common import GraphicUnitTest, UnitTestTouch
import networkx as nx

from LiSE import Engine
from LiSE.character import Facade
from ELiDE.app import ELiDEApp
from ELiDE.graph.board import GraphBoard, GraphBoardView, FinalLayout, \
 BoardScatterPlane
from .util import idle_until, window_with_widget, ELiDEAppTest
from ..dummy import Dummy


class GraphBoardTest(GraphicUnitTest):

	@staticmethod
	def test_layout_grid():
		spots_wide = 3
		spots_tall = 3
		graph = nx.grid_2d_graph(spots_wide, spots_tall)
		char = Facade(graph)
		app = ELiDEApp()
		spotlayout = FinalLayout()
		arrowlayout = FinalLayout()
		board = GraphBoard(app=app,
							character=char,
							spotlayout=spotlayout,
							arrowlayout=arrowlayout)
		spotlayout.pos = board.pos
		board.bind(pos=spotlayout.setter('pos'))
		spotlayout.size = board.size
		board.bind(size=spotlayout.setter('size'))
		board.add_widget(spotlayout)
		arrowlayout.pos = board.pos
		board.bind(pos=arrowlayout.setter('pos'))
		arrowlayout.size = board.size
		board.bind(size=arrowlayout.setter('size'))
		board.add_widget(arrowlayout)
		board.update()
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)

		@idle_until(timeout=1000, message="Never finished placing spots")
		def all_spots_placed():
			for x in range(spots_wide):
				for y in range(spots_tall):
					if (x, y) not in board.spot:
						return False
			return True

		# Don't get too picky about the exact proportions of the grid; just make sure the
		# spots are positioned logically with respect to one another
		for name, spot in board.spot.items():
			x, y = name
			if x > 0:
				assert spot.x > board.spot[x - 1, y].x
			if y > 0:
				assert spot.y > board.spot[x, y - 1].y
			if x < spots_wide - 1:
				assert spot.x < board.spot[x + 1, y].x
			if y < spots_tall - 1:
				assert spot.y < board.spot[x, y + 1].y

	@staticmethod
	def test_select_arrow():
		char = Facade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_place(1, _x=0.2, _y=0.1)
		char.add_portal(0, 1)
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(lambda: 0 in board.arrow and 1 in board.arrow[0] and board.
					arrow[0][1] in board.arrowlayout.children)
		ox, oy = board.spot[0].center
		dx, dy = board.spot[1].center
		motion = UnitTestTouch((ox + ((dx - ox) / 2)), dy)
		motion.touch_down()
		motion.touch_up()
		assert app.selection == board.arrow[0][1]

	@staticmethod
	def test_select_spot():
		char = Facade()
		char.add_place(0, _x=0.1, _y=0.1)
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(lambda: 0 in board.spot and board.spot[0] in board.
					spotlayout.children)
		x, y = board.spot[0].center
		motion = UnitTestTouch(x, y)
		motion.touch_down()
		motion.touch_up()
		assert app.selection == board.spot[0]

	@staticmethod
	def test_select_pawn():
		char = Facade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_thing('that', location=0)
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(
			lambda: 0 in board.spot and board.spot[0] in board.spotlayout.
			children and board.pawn['that'] in board.spot[0].children)
		motion = UnitTestTouch(*board.pawn['that'].center)
		motion.touch_down()
		motion.touch_up()
		assert app.selection == board.pawn['that']

	@staticmethod
	def test_pawn_relocate():
		char = Facade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_place(1, _x=0.2, _y=0.1)
		char.add_thing('that', location=0)
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(lambda: 0 in board.spot and board.spot[
			0] in board.spotlayout.children and 1 in board.spot and board.spot[
				1] in board.spotlayout.children and 'that' in board.pawn and
					board.pawn['that'] in board.spot[0].children)
		that = board.pawn['that']
		one = board.spot[1]
		# In a real ELiDE session, the following would happen as a
		# result of a Board.update() call
		char.thing['that']['location'] = that.loc_name = 1
		idle_until(lambda: that in one.children, 1000,
					"pawn did not relocate within 1000 ticks")

	def test_pawn_drag(self):
		char = Facade()
		char.add_place(0, _x=0.1, _y=0.1)
		char.add_place(1, _x=0.2, _y=0.1)
		char.add_thing('that', location=0)
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(lambda: 0 in board.spot and board.spot[
			0] in board.spotlayout.children and 1 in board.spot and board.spot[
				1] in board.spotlayout.children and 'that' in board.pawn and
					board.pawn['that'] in board.spot[0].children)
		that = board.pawn['that']
		one = board.spot[1]
		touch = UnitTestTouch(*that.center)
		touch.touch_down()
		dist_x = one.center_x - that.center_x
		dist_y = one.center_y - that.center_y
		for i in range(1, 11):
			coef = 1 / i
			x = one.center_x - coef * dist_x
			y = one.center_y - coef * dist_y
			touch.touch_move(x, y)
			self.advance_frames(1)
		that.pos = one.center
		self.advance_frames(1)
		touch.touch_up(*one.center)
		idle_until(lambda: that.pos != one.center, 100)
		idle_until(lambda: char.thing["that"]["location"] == 1, 100)

	@staticmethod
	def test_spot_and_pawn_from_dummy():
		char = Facade()
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		view = GraphBoardView(board=board)
		idle_until(lambda: view.plane is not None, 100,
					"Never made BoardScatterPlane")
		idle_until(lambda: board.spotlayout is not None, 100,
					"Never made SpotLayout")
		win = window_with_widget(view)
		dummy = Dummy(
			name='hello',
			paths=['atlas://rltiles/base/unseen'],
		)
		board.add_widget(dummy)
		idle_until(lambda: dummy in board.children, 100,
					"Dummy didn't get to board")
		dummy_name = dummy.name
		view.spot_from_dummy(dummy)
		idle_until(lambda: dummy_name in char.node, 100,
					"Dummy didn't add place")
		dummy2 = Dummy(name='goodbye',
						paths=['atlas://rltiles/base/unseen'],
						pos=dummy.center)
		dummy2_name = dummy2.name = 'dummy2'
		board.add_widget(dummy2)
		idle_until(lambda: dummy2 in board.children, 100,
					"Dummy 2 didn't get to board")
		view.pawn_from_dummy(dummy2)
		idle_until(lambda: dummy2_name in char.thing, 100,
					"Dummy 2 didn't add thing")
		idle_until(
			lambda: board.pawn[dummy2_name] in board.spot[dummy_name].children,
			100, "Dummy 2 didn't get to dummy 1")

	@staticmethod
	def test_pawn_add_new_place():
		char = Facade()
		app = ELiDEApp()
		board = GraphBoard(app=app, character=char)
		boardview = GraphBoardView(board=board)
		win = window_with_widget(boardview)
		idle_until(lambda: board.spotlayout)
		char.add_place(1, _x=0.2, _y=0.2)
		board.add_spot(1)
		idle_until(lambda: 1 in board.spot, 100, "Didn't make spot")
		char.add_thing('that', location=1)
		idle_until(lambda: 'that' in board.pawn, 100, "Didn't make pawn")
		that = board.pawn['that']
		one = board.spot[1]
		idle_until(lambda: that in one.children, 100,
					"pawn did not locate within 100 ticks")


class SwitchGraphTest(ELiDEAppTest):

	def test_character_switch_graph(self):
		with Engine(self.prefix) as eng:
			eng.add_character('physical', nx.grid_2d_graph(10, 1))
			eng.add_character('tall', nx.grid_2d_graph(1, 10))
		app = self.app
		window_with_widget(app.build())
		idle_until(
			lambda: hasattr(app, 'mainscreen') and app.mainscreen.mainview and
			app.mainscreen.statpanel and hasattr(app.mainscreen, 'gridview'))
		idle_until(lambda: app.mainscreen.boardview in app.mainscreen.mainview.
					children)
		idle_until(lambda: app.mainscreen.boardview.board.children)
		print(
			f'test_character_switch_graph got app {id(app)}, engine proxy {id(app.engine)}'
		)
		first_y = next(iter(
			app.mainscreen.boardview.board.spotlayout.children)).y
		assert all(
			child.y == first_y
			for child in app.mainscreen.boardview.board.spotlayout.children)
		assert len(
			set(child.x for child in
				app.mainscreen.boardview.board.spotlayout.children)) == len(
					app.mainscreen.boardview.board.spotlayout.children)
		app.character_name = 'tall'

		def all_x_same():
			if app.mainscreen.boardview.board is None or app.mainscreen.boardview.board.spotlayout is None or not app.mainscreen.boardview.board.spotlayout.children:
				return False
			first_x = next(
				iter(app.mainscreen.boardview.board.spotlayout.children)).x
			return all(child.x == first_x for child in
						app.mainscreen.boardview.board.spotlayout.children)

		idle_until(all_x_same, 1000, "Never got the new board")
		idle_until(
			lambda: len(
				set(child.y for child in app.mainscreen.boardview.board.
					spotlayout.children)
			) == len(app.mainscreen.boardview.board.spotlayout.children), 1000,
			"New board arranged weird")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Editors for textual data in the database.

The data is accessed via a "store" -- a mapping onto the table, used
like a dictionary. Each of the widgets defined here,
:class:`StringsEditor` and :class:`FuncsEditor`, displays a list of
buttons with which the user may select one of the keys in the store,
and edit its value in a text box.

"""
import re
import string
from functools import partial
from ast import parse
from textwrap import indent, dedent

from kivy.app import App
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.togglebutton import ToggleButton

from kivy.properties import (AliasProperty, BooleanProperty, ListProperty,
								NumericProperty, ObjectProperty,
								StringProperty)
from .util import trigger


class RecycleToggleButton(ToggleButton, RecycleDataViewBehavior):
	"""Toggle button at some index in a RecycleView"""
	index = NumericProperty()

	def on_touch_down(self, touch):
		if self.collide_point(*touch.pos):
			return self.parent.select_with_touch(self.index, touch)

	def apply_selection(self, rv, index, is_selected):
		if is_selected and index == self.index:
			self.state = 'down'
		else:
			self.state = 'normal'


class StoreButton(RecycleToggleButton):
	"""RecycleToggleButton to select something to edit in a Store"""
	store = ObjectProperty()
	"""Either a FunctionStore or a StringStore"""
	name = StringProperty()
	"""Name of this particular item"""
	source = StringProperty()
	"""Text of this item"""
	select = ObjectProperty()
	"""Function that gets called with my ``index`` when I'm selected"""

	def on_parent(self, *args):
		if self.name == '+':
			self.state = 'down'
			self.select(self.index)

	def on_state(self, *args):
		if self.state == 'down':
			self.select(self.index)


class StoreList(RecycleView):
	"""Holder for a :class:`kivy.uix.listview.ListView` that shows what's
    in a store, using one of the StoreAdapter classes.

    """
	store = ObjectProperty()
	"""Either a FunctionStore or a StringStore"""
	selection_name = StringProperty()
	"""The ``name`` of the ``StoreButton`` currently selected"""
	boxl = ObjectProperty()
	"""Instance of ``SelectableRecycleBoxLayout``"""

	def __init__(self, **kwargs):
		self._i2name = {}
		self._name2i = {}
		super().__init__(**kwargs)

	def on_store(self, *args):
		self.store.connect(self._trigger_redata)
		self.redata()

	def on_boxl(self, *args):
		self.boxl.bind(selected_nodes=self._pull_selection)

	def _pull_selection(self, *args):
		if not self.boxl.selected_nodes:
			return
		self.selection_name = self._i2name[self.boxl.selected_nodes[0]]

	def munge(self, datum):
		i, name = datum
		self._i2name[i] = name
		self._name2i[name] = i
		return {
			'store': self.store,
			'text': str(name),
			'name': name,
			'select': self.ids.boxl.select_node,
			'index': i
		}

	def _iter_keys(self):
		yield '+'
		yield from sorted(self.store._cache.keys())

	def redata(self, *args, **kwargs):
		"""Update my ``data`` to match what's in my ``store``"""
		select_name = kwargs.get('select_name')
		if not self.store:
			Clock.schedule_once(self.redata)
			return
		self.data = list(map(self.munge, enumerate(self._iter_keys())))
		if select_name:
			self._trigger_select_name(select_name)

	def _trigger_redata(self, *args, **kwargs):
		part = partial(self.redata, *args, **kwargs)
		if hasattr(self, '_scheduled_redata'):
			Clock.unschedule(self._scheduled_redata)
		self._scheduled_redata = Clock.schedule_once(part, 0)

	def select_name(self, name, *args):
		"""Select an item by its name, highlighting"""
		self.boxl.select_node(self._name2i[name])

	def _trigger_select_name(self, name):
		part = partial(self.select_name, name)
		if hasattr(self, '_scheduled_select_name'):
			Clock.unschedule(self._scheduled_select_name)
		self._scheduled_select_name = Clock.schedule_once(part, 0)


class LanguageInput(TextInput):
	"""Widget to enter the language you want to edit"""
	screen = ObjectProperty()
	"""The instance of ``StringsEdScreen`` that I'm in"""

	def on_focus(self, instance, value, *largs):
		if not value:
			if self.screen.language != self.text:
				self.screen.language = self.text
			self.text = ''


class StringsEdScreen(Screen):
	"""A screen in which to edit strings to be presented to humans

    Needs a ``toggle`` function to switch back to the main screen;
    a ``language`` identifier; and a ``language_setter`` function to be called
    with that ``language`` when changed.

    """
	toggle = ObjectProperty()
	"""Function to switch back to the main screen"""
	language = StringProperty('eng')
	"""Code identifying the language we're editing"""
	edbox = ObjectProperty()
	"""Widget containing editors for the current string and its name"""

	def on_language(self, *args):
		if self.edbox is None:
			Clock.schedule_once(self.on_language, 0)
			return
		self.edbox.storelist.redata()
		if self.store.language != self.language:
			self.store.language = self.language

	def on_store(self, *args):
		self.language = self.store.language
		self.store.language.connect(self._pull_language)

	def _pull_language(self, *args, language):
		self.language = language

	def save(self, *args):
		if self.edbox is None:
			Clock.schedule_once(self.save, 0)
			return
		self.edbox.save()


class Editor(BoxLayout):
	"""Abstract widget for editing strings or functions"""
	name_wid = ObjectProperty()
	"""Text input widget holding the name of the string being edited"""
	store = ObjectProperty()
	"""Proxy to the ``FunctionStore`` or ``StringStore``"""
	disable_text_input = BooleanProperty(False)
	"""Whether to prevent text entry (not name entry)"""
	# This next is the trigger on the EdBox, which may redata the StoreList
	_trigger_save = ObjectProperty()
	_trigger_delete = ObjectProperty()

	def save(self, *args):
		"""Put text in my store, return True if it changed"""
		if self.name_wid is None or self.store is None:
			Logger.debug("{}: Not saving, missing name_wid or store".format(
				type(self).__name__))
			return
		if not (self.name_wid.text or self.name_wid.hint_text):
			Logger.debug("{}: Not saving, no name".format(type(self).__name__))
			return
		if self.name_wid.text and self.name_wid.text[
			0] in string.digits + string.whitespace + string.punctuation:
			# TODO alert the user to invalid name
			Logger.warning("{}: Not saving, invalid name".format(
				type(self).__name__))
			return
		if hasattr(self, '_do_parse'):
			try:
				parse(self.source)
			except SyntaxError:
				# TODO alert user to invalid source
				Logger.debug("{}: Not saving, couldn't parse".format(
					type(self).__name__))
				return
		do_redata = False
		if self.name_wid.text:
			if (self.name_wid.hint_text
				and self.name_wid.hint_text != self.name_wid.text
				and hasattr(self.store, self.name_wid.hint_text)):
				delattr(self.store, self.name_wid.hint_text)
				do_redata = True
			if (not hasattr(self.store, self.name_wid.text)
				or getattr(self.store, self.name_wid.text) != self.source):
				Logger.debug("{}: Saving!".format(type(self).__name__))
				setattr(self.store, self.name_wid.text, self.source)
				do_redata = True
		elif self.name_wid.hint_text:
			if (not hasattr(self.store, self.name_wid.hint_text) or
				getattr(self.store, self.name_wid.hint_text) != self.source):
				Logger.debug("{}: Saving!".format(type(self).__name__))
				setattr(self.store, self.name_wid.hint_text, self.source)
				do_redata = True
		return do_redata

	def delete(self, *args):
		"""Remove the currently selected item from my store"""
		key = self.name_wid.text or self.name_wid.hint_text
		if not hasattr(self.store, key):
			# TODO feedback about missing key
			return
		delattr(self.store, key)
		try:
			return min(kee for kee in dir(self.store) if kee > key)
		except ValueError:
			return '+'


class StringInput(Editor):
	"""Editor for human-readable strings"""
	validate_name_input = ObjectProperty()
	"""Boolean function for checking if a string name is acceptable"""

	def on_name_wid(self, *args):
		if not self.validate_name_input:
			Clock.schedule_once(self.on_name_wid, 0)
			return
		self.name_wid.bind(text=self.validate_name_input)

	def _get_name(self):
		if self.name_wid:
			return self.name_wid.text

	def _set_name(self, v, *args):
		if not self.name_wid:
			Clock.schedule_once(partial(self._set_name, v), 0)
			return
		self.name_wid.text = v

	name = AliasProperty(_get_name, _set_name)

	def _get_source(self):
		if 'string' not in self.ids:
			return ''
		return self.ids.string.text

	def _set_source(self, v, *args):
		if 'string' not in self.ids:
			Clock.schedule_once(partial(self._set_source, v), 0)
			return
		self.ids.string.text = v

	source = AliasProperty(_get_source, _set_source)


class EdBox(BoxLayout):
	"""Box containing most of an editor's screen

    Has a StoreList and an Editor, which in turn holds a name field and a big text entry box.

    """
	storelist = ObjectProperty()
	"""An instance of ``StoreList``"""
	editor = ObjectProperty()
	"""An instance of a subclass of ``Editor``"""
	store = ObjectProperty()
	"""Proxy to the store I represent"""
	store_name = StringProperty()
	"""Name of my store, so I can get it from the engine"""
	data = ListProperty()
	"""Dictionaries describing widgets in my ``storelist``"""
	toggle = ObjectProperty()
	"""Function to show or hide my screen"""
	disable_text_input = BooleanProperty(False)
	"""Set to ``True`` to prevent entering text in the editor"""

	def on_store_name(self, *args):
		app = App.get_running_app()
		if app.engine is None:
			Clock.schedule_once(self.on_store_name, 0)
			return
		self.store = getattr(app.engine, self.store_name)

	def on_storelist(self, *args):
		self.storelist.bind(selection_name=self._pull_from_storelist)

	@trigger
	def validate_name_input(self, *args):
		self.disable_text_input = not (
			self.valid_name(self.editor.name_wid.hint_text)
			or self.valid_name(self.editor.name_wid.text))

	@trigger
	def _pull_from_storelist(self, *args):
		self.save()
		# The + button at the top is for adding an entry yet unnamed, so don't display hint text for it
		self.editor.name_wid.hint_text = self.storelist.selection_name.strip(
			'+')
		self.editor.name_wid.text = ''
		try:
			self.editor.source = getattr(self.store,
											self.editor.name_wid.hint_text)
		except AttributeError:
			self.editor.source = self.get_default_text(
				self.editor.name_wid.hint_text)
		self.disable_text_input = not self.valid_name(
			self.editor.name_wid.hint_text)
		if hasattr(self, '_lock_save'):
			del self._lock_save

	def dismiss(self, *args):
		self.save()
		self.toggle()

	def save(self, *args, name=None):
		if not self.editor or not self.store:
			return
		if hasattr(self, '_lock_save'):
			return
		self._lock_save = True
		save_select = self.editor.save()
		if save_select:
			self.storelist.redata(select_name=name)
		else:
			del self._lock_save

	def _trigger_save(self, name=None):
		part = partial(self.save, name=name)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	def delete(self, *args):
		if not self.editor:
			return
		if hasattr(self, '_lock_save'):
			return
		self._lock_save = True
		del_select = self.editor.delete()
		if del_select:
			self.storelist.redata(del_select)
		else:
			del self._lock_save

	_trigger_delete = trigger(delete)

	def on_store(self, *args):
		pass


class StringNameInput(TextInput):
	"""Small text box for the names of strings"""
	_trigger_save = ObjectProperty()

	def on_focus(self, inst, val, *largs):
		if self.text and not val:
			self._trigger_save(self.text)


class StringsEdBox(EdBox):
	"""Box containing most of the strings editing screen

    Contains the storelist and the editor, which in turn contains the string name input
    and a bigger input field for the string itself.

    """
	language = StringProperty('eng')

	@staticmethod
	def get_default_text(newname):
		return ''

	@staticmethod
	def valid_name(name):
		return name and name[0] != '+'


sig_ex = re.compile(r'^ *def .+?\((.+)\):$')


class FunctionNameInput(TextInput):
	"""Input for the name of a function

    Filters out illegal characters.

    """
	_trigger_save = ObjectProperty()

	def insert_text(self, s, from_undo=False):
		if self.text == '':
			if s[0] not in (string.ascii_letters + '_'):
				return
		return super().insert_text(''.join(
			c for c in s if c in (string.ascii_letters + string.digits + '_')))

	def on_focus(self, inst, val, *largs):
		if not val:
			self._trigger_save(self.text)


def munge_source(v):
	"""Take Python source code, return a pair of its parameters and the rest of it dedented"""
	lines = v.split('\n')
	if not lines:
		return tuple(), ''
	firstline = lines[0].lstrip()
	while firstline == '' or firstline[0] == '@':
		del lines[0]
		firstline = lines[0].lstrip()
	if not lines:
		return tuple(), ''
	params = tuple(parm.strip()
					for parm in sig_ex.match(lines[0]).group(1).split(','))
	del lines[0]
	if not lines:
		return params, ''
	# hack to allow 'empty' functions
	if lines and lines[-1].strip() == 'pass':
		del lines[-1]
	return params, dedent('\n'.join(lines))


class FuncEditor(Editor):
	"""The editor widget for working with any particular function.

    Contains a one-line field for the function's name and a multi-line
    field for its code.

    """
	storelist = ObjectProperty()
	"""Instance of ``StoreList`` that shows all the functions you can edit"""
	codeinput = ObjectProperty()
	params = ListProperty(['obj'])
	_text = StringProperty()
	_do_parse = True

	def _get_source(self):
		code = self.get_default_text(self.name_wid.text
										or self.name_wid.hint_text)
		if self._text:
			code += indent(self._text, ' ' * 4)
		else:
			code += ' ' * 4 + 'pass'
		return code.rstrip(' \n\t')

	def _set_source(self, v):
		if not self.codeinput:
			Clock.schedule_once(partial(self._set_source, v), 0)
			return
		self.codeinput.unbind(text=self.setter('_text'))
		self.params, self.codeinput.text = munge_source(str(v))
		self.codeinput.bind(text=self.setter('_text'))

	source = AliasProperty(_get_source, _set_source, bind=('_text', 'params'))

	def get_default_text(self, name):
		if not name or name == '+':
			name = 'a'
		return "def {}({}):\n".format(name, ', '.join(self.params))

	def on_codeinput(self, *args):
		self._text = self.codeinput.text
		self.codeinput.bind(text=self.setter('_text'))


class FuncsEdBox(EdBox):
	"""Widget for editing the Python source of funcs to be used in LiSE sims.

    Contains a list of functions in the store it's about, next to a
    FuncEditor showing the source of the selected one, and a close button.

    """

	def get_default_text(self, newname):
		return self.editor.get_default_text(newname)

	@staticmethod
	def valid_name(name):
		return name and name[
			0] not in string.digits + string.whitespace + string.punctuation


class FuncsEdScreen(Screen):
	"""Screen containing three FuncsEdBox

    Triggers, prereqs, and actions.

    """
	toggle = ObjectProperty()

	def save(self, *args):
		self.ids.triggers.save()
		self.ids.prereqs.save()
		self.ids.actions.save()


Builder.load_string("""
#: import py3lexer pygments.lexers.Python3Lexer
<StoreButton>:
    size_hint_y: None
    height: 30
<StoreList>:
    viewclass: 'StoreButton'
    boxl: boxl
    SelectableRecycleBoxLayout:
        id: boxl
        default_size: None, dp(56)
        default_size_hint: 1, None
        height: self.minimum_height
        size_hint_y: None
        orientation: 'vertical'
<StringInput>:
    name_wid: stringname
    orientation: 'vertical'
    BoxLayout:
        size_hint_y: 0.05
        Label:
            id: title
            text: 'Title: '
            size_hint_x: None
            width: self.texture_size[0]
        StringNameInput:
            id: stringname
            multiline: False
            write_tab: False
            _trigger_save: root._trigger_save
        Button:
            text: 'del'
            size_hint_x: 0.1
            on_release: root._trigger_delete()
    TextInput:
        id: string
        disabled: root.disable_text_input
<StringsEdBox>:
    editor: strings_ed
    storelist: strings_list
    orientation: 'vertical'
    BoxLayout:
        orientation: 'horizontal'
        StoreList:
            id: strings_list
            size_hint_x: 0.2
            store: root.store
        StringInput:
            id: strings_ed
            store: root.store
            disable_text_input: root.disable_text_input
            validate_name_input: root.validate_name_input
            _trigger_save: root._trigger_save
            _trigger_delete: root._trigger_delete
<StringsEdScreen>:
    name: 'strings'
    edbox: edbox
    BoxLayout:
        orientation: 'vertical'
        StringsEdBox:
            id: edbox
            toggle: root.toggle
            store_name: 'string'
            language: root.language
        BoxLayout:
            size_hint_y: 0.05
            Button:
                text: 'Close'
                on_release: edbox.dismiss()
            Label:
                text_size: self.size
                halign: 'right'
                valign: 'middle'
                text: 'Language: '
            LanguageInput:
                id: language
                screen: root
                hint_text: root.language
                write_tab: False
                multiline: False
<Py3CodeInput@CodeInput>:
    lexer: py3lexer()
<FuncEditor>:
    codeinput: code
    name_wid: funname
    orientation: 'vertical'
    BoxLayout:
        orientation: 'horizontal'
        size_hint_y: None
        height: funname.height
        Py3CodeInput:
            id: imafunction
            text: 'def'
            disabled: True
            size_hint: (None, None)
            height: self.line_height + self.font_size
            width: self.font_size * len(self.text)
            background_disabled_normal: ''
            disabled_foreground_color: self.foreground_color
        FunctionNameInput:
            id: funname
            size_hint: (0.4, None)
            height: self.line_height + self.font_size
            multiline: False
            write_tab: False
            _trigger_save: root._trigger_save
            on_text: root.validate_name_input(self.text)
        Py3CodeInput:
            id: params
            text: '({}):'.format(', '.join(root.params))
            disabled: True
            size_hint_y: None
            height: self.line_height + self.font_size
            background_disabled_normal: ''
            disabled_foreground_color: self.foreground_color
        Button:
            text: 'del'
            size_hint_x: 0.1
            on_release: root._trigger_delete()
    BoxLayout:
        orientation: 'horizontal'
        Label:
            canvas:
                Color:
                    rgba: params.background_color
                Rectangle:
                    pos: self.pos
                    size: self.size
                Color:
                    rgba: [1., 1., 1., 1.]
            # PEP8 standard indentation width is 4 spaces
            text: ' ' * 4
            size_hint_x: None
            width: self.texture_size[0]
        Py3CodeInput:
            id: code
            disabled: root.disable_text_input
<FuncsEdBox>:
    editor: funcs_ed
    storelist: funcs_list
    orientation: 'vertical'
    data: [(item['name'], self.store.get_source(item['name'])) for item in funcs_list.data[1:]]
    BoxLayout:
        orientation: 'horizontal'
        BoxLayout:
            orientation: 'vertical'
            size_hint_x: 0.2
            StoreList:
                id: funcs_list
                store: root.store
        FuncEditor:
            id: funcs_ed
            store: root.store
            storelist: funcs_list
            disable_text_input: root.disable_text_input
            validate_name_input: root.validate_name_input
            _trigger_save: root._trigger_save
            _trigger_delete: root._trigger_delete
    BoxLayout:
        size_hint_y: 0.05
        Button:
            text: 'Close'
            on_release: root.dismiss()
            size_hint_x: 0.2
        Widget:
            id: spacer
<FuncsEdScreen>:
    name: 'funcs'
    TabbedPanel:
        default_tab: action
        TabbedPanelItem:
            id: trigger
            text: 'Trigger'
            on_state: triggers.save()
            FuncsEdBox:
                id: triggers
                toggle: root.toggle
                store_name: 'trigger'
                on_data: app.rules.rulesview.set_functions('trigger', map(app.rules.rulesview.inspect_func, self.data))
        TabbedPanelItem:
            id: prereq
            text: 'Prereq'
            on_state: prereqs.save()
            FuncsEdBox:
                id: prereqs
                toggle: root.toggle
                store_name: 'prereq'
                on_data: app.rules.rulesview.set_functions('prereq', map(app.rules.rulesview.inspect_func, self.data))
        TabbedPanelItem:
            id: action
            text: 'Action'
            on_state: actions.save()
            FuncsEdBox:
                id: actions
                toggle: root.toggle
                store_name: 'action'
                on_data: app.rules.rulesview.set_functions('action', map(app.rules.rulesview.inspect_func, self.data))
""")



from functools import partial

from kivy.clock import Clock
from kivy.logger import Logger
from kivy.properties import (DictProperty, ListProperty, NumericProperty,
								ObjectProperty, ReferenceListProperty)
from kivy.uix.relativelayout import RelativeLayout
from kivy.lang.builder import Builder
from .spot import GridSpot
from .pawn import GridPawn
from ..boardview import BoardView


class GridBoard(RelativeLayout):
	selection = ObjectProperty()
	selection_candidates = ListProperty()
	character = ObjectProperty()
	tile_width = NumericProperty()
	tile_height = NumericProperty()
	tile_size = ReferenceListProperty(tile_width, tile_height)
	pawn = DictProperty({})
	spot = DictProperty({})
	spot_cls = ObjectProperty(GridSpot)
	pawn_cls = ObjectProperty(GridPawn)
	_parented: bool

	def do_layout(self, *args):
		Logger.debug("GridBoard laying out at size {}".format(self.size))
		rows = cols = 0
		colw = rowh = 0
		for name, spot in self.spot.items():
			row, col = name
			rows = max((rows, row))
			cols = max((cols, col))
			colw = max((spot.width, colw))
			rowh = max((spot.height, rowh))
		self.width = colw * cols
		self.height = rowh * rows
		for name, tile in self.spot.items():
			gx, gy = name
			tile.pos = gx * colw, gy * rowh

	def add_spot(self, placen, *args):
		if placen not in self.character.place:
			raise KeyError(f"No such place for spot: {placen}")
		self.add_widget(self.make_spot(self.character.place[placen]))

	def make_spot(self, place):
		placen = place["name"]
		if placen in self.spot:
			raise KeyError("Already have a Spot for this Place")
		if not isinstance(placen, tuple) or len(placen) != 2 or not isinstance(
			placen[0], int) or not isinstance(placen[1], int):
			raise TypeError(
				"GridBoard can only display places named with pairs of ints")

		r = self.spot_cls(board=self, proxy=place)
		self.spot[place["name"]] = r
		return r

	def make_pawn(self, thing):
		if thing["name"] in self.pawn:
			raise KeyError("Already have a Pawn for this Thing")
		r = self.pawn_cls(board=self, proxy=thing)
		self.pawn[thing["name"]] = r
		return r

	def _trigger_add_tile(self, placen):
		part = partial(self.add_spot, placen)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	def add_new_spots(self, *args):
		return self._maybe_add_nodes(GridSpot.default_image_paths,
										self.character.place, self.spot,
										self.make_spot, self.add_widget)

	def add_pawn(self, thingn, *args):
		if thingn not in self.character.thing:
			raise KeyError(f"No such thing: {thingn}")
		if thingn in self.pawn:
			raise KeyError(f"Already have a pawn for {thingn}")
		pwn = self.make_pawn(self.character.thing[thingn])
		whereat = self.spot[pwn.proxy['location']]
		whereat.add_widget(pwn)
		self.pawn[thingn] = pwn

	def _trigger_add_pawn(self, thingn):
		part = partial(self.add_pawn, thingn)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)

	@staticmethod
	def _maybe_add_nodes(default_image_paths, nodemap, mymap, maker, adder):
		default_zeroes = [0] * len(default_image_paths)
		to_add = []
		patchmap = {}
		for noden, node in nodemap.items():
			if noden not in mymap:
				to_add.append(node)
				patch = {}
				if '_image_paths' in node:
					zeroes = [0] * len(node['_image_paths'])
				else:
					patch['_image_paths'] = default_image_paths
					zeroes = default_zeroes
				if '_offxs' not in node:
					patch['_offxs'] = zeroes
				if '_offys' not in node:
					patch['_offys'] = zeroes
				if patch:
					patchmap[noden] = patch
		if patchmap:
			nodemap.patch(patchmap)
		added = []
		for node in to_add:
			try:
				mine = maker(node)
			except TypeError:
				continue
			added.append(mine)
			adder(mine)
		return added

	def add_new_pawns(self, *args):

		def adder(pwn):
			self.spot[pwn.proxy['location']].add_widget(pwn)

		return self._maybe_add_nodes(GridPawn.default_image_paths,
										self.character.thing, self.pawn,
										self.make_pawn, adder)

	def on_parent(self, *args):
		if not self.parent or hasattr(self, '_parented'):
			return
		self._parented = True
		self.update()

	def remove_absent_pawns(self):
		pawnmap = self.pawn
		thingmap = self.character.thing
		for name, pawn in list(pawnmap.items()):
			if name not in thingmap:
				pawn.parent.remove_widget(pawn)
				del pawnmap[name]

	def remove_absent_spots(self):
		spotmap = self.spot
		placemap = self.character.place
		remove_widget = self.remove_widget
		for name, spot in list(spotmap.items()):
			if name not in placemap:
				remove_widget(spot)
				del spotmap[name]

	def update(self, *args):
		self.remove_absent_pawns()
		self.remove_absent_spots()
		self.add_new_spots()
		self.add_new_pawns()
		for spot in self.spot.values():
			spot.finalize()
		for pawn in self.pawn.values():
			pawn.finalize()

	def rm_spot(self, name):
		spot = self.spot.pop(name)
		if spot in self.selection_candidates:
			self.selection_candidates.remove(spot)
		for pwn in spot.children:
			del self.pawn[pwn.name]
		self.remove_widget(spot)

	def rm_pawn(self, name):
		pwn = self.pawn.pop(name)
		if pwn in self.selection_candidates:
			self.selection_candidates.remove(pwn)
		pwn.parent.remove_widget(pwn)

	def update_from_delta(self, delta, *args):
		pawnmap = self.pawn
		spotmap = self.spot
		add_pawn = self.add_pawn
		add_spot = self.add_spot
		selection_candidates = self.selection_candidates

		def rm_pawn(name):
			pwn = pawnmap.pop(name)
			if pwn in selection_candidates:
				selection_candidates.remove(pwn)
			pwn.parent.remove_widget(pwn)

		remove_widget = self.remove_widget

		def rm_spot(name):
			spot = spotmap.pop(name)
			if spot in selection_candidates:
				selection_candidates.remove(spot)
			for pwn in spot.children:
				del pawnmap[pwn.name]
			remove_widget(spot)

		if 'nodes' in delta:
			for node, extant in delta['nodes'].items():
				if extant:
					if 'node_val' in delta \
                                     and node in delta['node_val'] \
                                     and 'location' in delta['node_val'][node] \
                                     and node not in pawnmap:
						add_pawn(node)
					elif node not in spotmap:
						add_spot(node)
				else:
					if node in pawnmap:
						rm_pawn(node)
					if node in spotmap:
						rm_spot(node)
		if 'node_val' in delta:
			for node, stats in delta['node_val'].items():
				if node in spotmap and '_image_paths' in stats:
					spotmap[node].paths = stats['_image_paths'] \
                                                         or GridSpot.default_image_paths
				elif node in pawnmap:
					pawn = pawnmap[node]
					if 'location' in stats:
						pawn.loc_name = stats['location']
					if '_image_paths' in stats:
						pawn.paths = stats['_image_paths'] \
                                                       or GridPawn.default_image_paths
				else:
					Logger.warning(
						"GridBoard: diff tried to change stats of node {}"
						"but I don't have a widget for it".format(node))

	def trigger_update_from_delta(self, delta, *args):
		part = partial(self.update_from_delta, delta)
		Clock.unschedule(part)
		Clock.schedule_once(part, 0)


class GridBoardView(BoardView):
	pass


Builder.load_string("""
<GridBoard>:
    app: app
    size_hint: None, None
<GridBoardView>:
    plane: boardplane
    BoardScatterPlane:
        id: boardplane
        board: root.board
        scale_min: root.scale_min
        scale_max: root.scale_max
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock
from kivy.lang import Builder
from kivy.uix.boxlayout import BoxLayout

from kivy.properties import (BooleanProperty, ObjectProperty,
								ReferenceListProperty)
from .graph.arrow import GraphArrowWidget
from .util import try_load, dummynum
from LiSE.proxy import CharStatProxy


class CharMenu(BoxLayout):
	screen = ObjectProperty()
	reciprocal_portal = BooleanProperty(True)
	revarrow = ObjectProperty(None, allownone=True)
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	dummies = ReferenceListProperty(dummyplace, dummything)

	@property
	def app(self):
		if not self.screen:
			raise AttributeError("No screen, therefore no app")
		return self.screen.app

	@property
	def engine(self):
		if not self.screen or not self.screen.app:
			raise AttributeError("Can't get engine from screen")
		return self.screen.app.engine

	def on_screen(self, *args):
		if not (self.screen and self.screen.boardview and self.screen.app):
			Clock.schedule_once(self.on_screen, 0)
			return
		self.screen.boardview.reciprocal_portal = self.reciprocal_portal
		if self.reciprocal_portal:
			assert (self.revarrow is None)
			self.revarrow = GraphArrowWidget(board=self.screen.boardview.board,
												origin=self.ids.emptyright,
												destination=self.ids.emptyleft)
			self.ids.portaladdbut.add_widget(self.revarrow)
		self.bind(reciprocal_portal=self.screen.boardview.setter(
			'reciprocal_portal'))

	def spot_from_dummy(self, dummy):
		if self.screen.boardview.parent != self.screen.mainview:
			return
		name = dummy.name
		self.screen.boardview.spot_from_dummy(dummy)
		graphboard = self.screen.graphboards[self.app.character_name]
		if name not in graphboard.spot:
			graphboard.add_spot(name)
		gridboard = self.screen.gridboards[self.app.character_name]
		if name not in gridboard.spot and isinstance(
			name, tuple) and len(name) == 2:
			gridboard.add_spot(name)

	def pawn_from_dummy(self, dummy):
		name = dummy.name
		self.screen.mainview.children[0].pawn_from_dummy(dummy)
		graphboard = self.screen.graphboards[self.app.character_name]
		if name not in graphboard.pawn:
			graphboard.add_pawn(name)
		gridboard = self.screen.gridboards[self.app.character_name]
		if name not in gridboard.pawn and self.app.character.thing[name][
			"location"] in gridboard.spot:
			gridboard.add_pawn(name)

	def toggle_chars_screen(self, *args):
		"""Display or hide the list you use to switch between characters."""
		# TODO: update the list of chars
		self.app.chars.toggle()

	def toggle_rules(self, *args):
		"""Display or hide the view for constructing rules out of cards."""
		if self.app.manager.current != 'rules' and not isinstance(
			self.app.selected_proxy, CharStatProxy):
			self.app.rules.entity = self.app.selected_proxy
			self.app.rules.rulebook = self.app.selected_proxy.rulebook
		if isinstance(self.app.selected_proxy, CharStatProxy):
			self.app.charrules.character = self.app.selected_proxy
			self.app.charrules.toggle()
		else:
			self.app.rules.toggle()

	def toggle_funcs_editor(self):
		"""Display or hide the text editing window for functions."""
		self.app.funcs.toggle()

	def toggle_strings_editor(self):
		self.app.strings.toggle()

	def toggle_spot_cfg(self):
		"""Show the dialog where you select graphics and a name for a place,
        or hide it if already showing.

        """
		if self.app.manager.current == 'spotcfg':
			dummyplace = self.screendummyplace
			self.ids.placetab.remove_widget(dummyplace)
			dummyplace.clear()
			if self.app.spotcfg.prefix:
				dummyplace.prefix = self.app.spotcfg.prefix
				dummyplace.num = dummynum(self.app.character,
											dummyplace.prefix) + 1
			if self.app.spotcfg.imgpaths:
				dummyplace.paths = self.app.spotcfg.imgpaths
			else:
				dummyplace.paths = ['atlas://rltiles/floor/floor-stone']
			dummyplace.center = self.ids.placetab.center
			self.ids.placetab.add_widget(dummyplace)
		else:
			self.app.spotcfg.prefix = self.ids.dummyplace.prefix
		self.app.spotcfg.toggle()

	def toggle_pawn_cfg(self):
		"""Show or hide the pop-over where you can configure the dummy pawn"""
		if self.app.manager.current == 'pawncfg':
			dummything = self.app.dummything
			self.ids.thingtab.remove_widget(dummything)
			dummything.clear()
			if self.app.pawncfg.prefix:
				dummything.prefix = self.app.pawncfg.prefix
				dummything.num = dummynum(self.app.character,
											dummything.prefix) + 1
			if self.app.pawncfg.imgpaths:
				dummything.paths = self.app.pawncfg.imgpaths
			else:
				dummything.paths = ['atlas://rltiles/base/unseen']
			self.ids.thingtab.add_widget(dummything)
		else:
			self.app.pawncfg.prefix = self.ids.dummything.prefix
		self.app.pawncfg.toggle()

	def toggle_reciprocal(self):
		"""Flip my ``reciprocal_portal`` boolean, and draw (or stop drawing)
        an extra arrow on the appropriate button to indicate the
        fact.

        """
		self.screen.boardview.reciprocal_portal = not self.screen.boardview.reciprocal_portal
		if self.screen.boardview.reciprocal_portal:
			assert (self.revarrow is None)
			self.revarrow = GraphArrowWidget(board=self.screen.boardview.board,
												origin=self.ids.emptyright,
												destination=self.ids.emptyleft)
			self.ids.portaladdbut.add_widget(self.revarrow)
		else:
			if hasattr(self, 'revarrow'):
				self.ids.portaladdbut.remove_widget(self.revarrow)
				self.revarrow = None

	def new_character(self, but):
		charn = try_load(self.app.engine.unpack,
							self.app.chars.ids.newname.text)
		self.app.select_character(self.app.engine.new_character(charn))
		self.app.chars.ids.newname.text = ''
		self.app.chars.charsview.adapter.data = list(
			self.engine.character.keys())
		Clock.schedule_once(self.toggle_chars_screen, 0.01)

	def on_dummyplace(self, *args):
		if not self.dummyplace.paths:
			self.dummyplace.paths = ["atlas://rltiles/floor.atlas/floor-stone"]

	def on_dummything(self, *args):
		if not self.dummything.paths:
			self.dummything.paths = ["atlas://rltiles/base.atlas/unseen"]


Builder.load_string("""
<CharMenu>:
    orientation: 'vertical'
    dummyplace: dummyplace
    dummything: dummything
    portaladdbut: portaladdbut
    portaldirbut: portaldirbut
    Button:
        text: 'Characters'
        on_release: root.toggle_chars_screen()
    Button:
        text: 'Strings'
        on_release: root.toggle_strings_editor()
    Button:
        text: 'Python'
        on_release: root.toggle_funcs_editor()
    Button:
        text: 'Rules'
        on_release: root.toggle_rules()
    Button:
        text: 'Delete'
        on_release: app.delete_selection()
    BoxLayout:
        Widget:
            id: placetab
            Dummy:
                id: dummyplace
                center: placetab.center
                prefix: 'place'
                on_pos_up: root.spot_from_dummy(self)
        Button:
            text: 'cfg'
            on_release: root.toggle_spot_cfg()
    BoxLayout:
        orientation: 'vertical'
        ToggleButton:
            id: portaladdbut
            Widget:
                id: emptyleft
                center_x: portaladdbut.x + portaladdbut.width / 3
                center_y: portaladdbut.center_y
                size: (0, 0)
            Widget:
                id: emptyright
                center_x: portaladdbut.right - portaladdbut.width / 3
                center_y: portaladdbut.center_y
                size: (0, 0)
            GraphArrowWidget:
                graph: root.screen.boardview.graph if root.screen and root.screen.boardview else None
                origin: emptyleft
                destination: emptyright
        Button:
            id: portaldirbut
            text: 'One-way' if root.reciprocal_portal else 'Two-way'
            on_release: root.toggle_reciprocal()
    BoxLayout:
        Widget:
            id: thingtab
            Dummy:
                id: dummything
                center: thingtab.center
                prefix: 'thing'
                on_pos_up: root.pawn_from_dummy(self)
        Button:
            text: 'cfg'
            on_release: root.toggle_pawn_cfg()
""")
